import logging
import os

from cekit.config import Config
from cekit.generator.base import Generator

logger = logging.getLogger("cekit")
config = Config()


class DockerGenerator(Generator):
    def __init__(self, descriptor_path, target, overrides):
        super(DockerGenerator, self).__init__(descriptor_path, target, overrides)
        self._fetch_repos = True

    def prepare_artifacts(self):
        """Goes through artifacts section of image descriptor
        and fetches all of them
        """

        logger.info("Handling artifacts for docker...")
        target_dir = os.path.join(self.target, "image")

        for image in self.images:
            for artifact in image.all_artifacts:
                artifact.copy(target_dir)

        logger.debug("Artifacts handled")



# flake8: noqa
# forwarding import
from cekit.descriptor.base import Descriptor
from cekit.descriptor.resource import Resource
from cekit.descriptor.label import Label
from cekit.descriptor.port import Port
from cekit.descriptor.env import Env
from cekit.descriptor.volume import Volume
from cekit.descriptor.osbs import Osbs
from cekit.descriptor.run import Run
from cekit.descriptor.packages import Packages
from cekit.descriptor.packages import Repository
from cekit.descriptor.modules import Modules
from cekit.descriptor.execute import Execute
from cekit.descriptor.image import Image
from cekit.descriptor.module import Module
from cekit.descriptor.overrides import Overrides



import logging

import yaml

import cekit
from cekit.descriptor import Descriptor

execute_schemas = yaml.safe_load(
    """
        map:
          name: {type: str}
          script: {type: str}
          user: {type: text}"""
)

container_schemas = yaml.safe_load(
    """
        seq:
          - {type: any}"""
)

logger = logging.getLogger("cekit")


class Execute(Descriptor):
    def __init__(self, descriptor, module_name):
        self.schema = execute_schemas
        super(Execute, self).__init__(descriptor)

        descriptor["directory"] = module_name
        descriptor["module_name"] = module_name

        if "name" not in descriptor:
            descriptor["name"] = "{}/{}".format(module_name, descriptor["script"])

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def script(self):
        return self.get("script")

    @script.setter
    def script(self, value):
        self._descriptor["script"] = value

    @property
    def user(self):
        return self.get("user", cekit.DEFAULT_USER)

    @user.setter
    def user(self, value):
        self._descriptor["user"] = value



# -*- coding: utf-8 -*-

import logging
import os
import platform
import re
import shutil
import tempfile
from urllib.parse import urlparse

import yaml

from jinja2 import Environment, FileSystemLoader
from packaging.version import LegacyVersion
from packaging.version import parse as parse_version

from cekit.config import Config
from cekit.descriptor import Env, Image, Label, Module, Overrides, Repository
from cekit.errors import CekitError
from cekit.template_helper import TemplateHelper
from cekit.tools import download_file, load_descriptor
from cekit.version import __version__ as cekit_version

LOGGER = logging.getLogger("cekit")
CONFIG = Config()

try:
    # Requests is a dependency of ODCS client, so this should be safe
    import requests
    from odcs.client.odcs import ODCS, AuthMech
except ImportError:
    pass


class Generator(object):
    """
    This class process Image descriptor(self.image) and uses it to generate
    target directory by fetching all dependencies and artifacts

    Args:
      descriptor_path - path to an image descriptor
      target - path to target directory
      overrides - path to overrides file (can be None)
    """

    ODCS_HIDDEN_REPOS_FLAG = "include_unpublished_pulp_repos"

    def __init__(self, descriptor_path, target, overrides):
        self._descriptor_path = descriptor_path
        self._overrides = []
        self.target = target
        self._fetch_repos = False
        self._module_registry = ModuleRegistry()
        self.image = None
        self.builder_images = []
        self.images = []

        # If descriptor has been passed in from standard input its not a path so use current working directory
        if "-" == descriptor_path:
            descriptor_path = os.getcwd()

        if overrides:
            for override in overrides:

                LOGGER.debug("Loading override '{}'".format(override))
                if urlparse(override).scheme in ["http", "https", "file"]:
                    # HTTP Handling
                    tmpfile = tempfile.NamedTemporaryFile()
                    download_file(override, tmpfile.name)
                    self._overrides.append(
                        Overrides(
                            load_descriptor(tmpfile.name), os.path.dirname(tmpfile.name)
                        )
                    )
                else:
                    # File handling
                    override_artifact_dir = os.path.dirname(os.path.abspath(override))
                    if not os.path.exists(override):
                        override_artifact_dir = os.path.dirname(
                            os.path.abspath(descriptor_path)
                        )
                    self._overrides.append(
                        Overrides(load_descriptor(override), override_artifact_dir)
                    )

        LOGGER.info("Initializing image descriptor...")

    @staticmethod
    def dependencies(params=None):
        deps = {}

        deps["odcs-client"] = {
            "package": "python3-odcs-client",
            "library": "odcs",
        }

        if CONFIG.get("common", "redhat"):
            deps["brew"] = {"package": "brewkoji", "executable": "/usr/bin/brew"}
            # FOLLOW_TAG requires both version and release labels ; these are generally
            # only added consistently within RH and Fedora repos so currently this annotation
            # is only supported under the RH flag.
            deps["skopeo"] = {"package": "skopeo", "executable": "/usr/bin/skopeo"}

        return deps

    def init(self):
        """
        Initializes the image object.
        """

        LOGGER.debug("Removing old target directory")
        shutil.rmtree(self.target, ignore_errors=True)
        os.makedirs(os.path.join(self.target, "image"))

        # Read the main image descriptor and create an Image object from it
        descriptor = load_descriptor(self._descriptor_path)

        if isinstance(descriptor, list):
            LOGGER.info(
                "Descriptor contains multiple elements, assuming multi-stage image"
            )
            LOGGER.info(
                "Found {} builder image(s) and one target image".format(
                    len(descriptor[:-1])
                )
            )

            # Iterate over images defined in image descriptor and
            # create Image objects out of them
            for image_descriptor in descriptor[:-1]:
                self.builder_images.append(
                    Image(
                        image_descriptor,
                        os.path.dirname(os.path.abspath(self._descriptor_path)),
                    )
                )

            descriptor = descriptor[-1]

        self.image = Image(
            descriptor, os.path.dirname(os.path.abspath(self._descriptor_path))
        )

        # Construct list of all images (builder images + main one)
        self.images = [self.image] + self.builder_images

        for image in self.images:
            # Apply overrides to all image definitions:
            # intermediate (builder) images and target image as well
            # It is required to build the module registry
            image.apply_image_overrides(self._overrides)

        # Load definitions of modules
        # We need to load it after we apply overrides so that any changes to modules
        # will be reflected there as well
        self.build_module_registry()

        for image in self.images:
            # Process included modules
            image.apply_module_overrides(self._module_registry)
            image.process_defaults()

        # Add build labels
        self.add_build_labels()

    def generate(self):
        self.copy_modules()
        self.prepare_artifacts()
        self.prepare_repositories()
        self.image.remove_none_keys()
        self.image.write(os.path.join(self.target, "image.yaml"))
        self.render_dockerfile()
        self.render_help()

    def add_redhat_overrides(self):
        self._overrides.append(self.get_redhat_overrides())

    def add_build_labels(self):
        image_labels = self.image.labels
        # we will persist cekit version in a label here, so we know which version of cekit
        # was used to build the image
        image_labels.append(Label({"name": "io.cekit.version", "value": cekit_version}))

        for label in image_labels:
            if len(label.value) > 128:
                # breaks the line each time it reaches 128 characters
                label.value = "\\\n".join(re.findall("(?s).{,128}", label.value))[:]

        # If we define the label in the image descriptor
        # we should *not* override it with value from
        # the root's key
        if self.image.description and not self.image.label("description"):
            image_labels.append(
                Label({"name": "description", "value": self.image.description})
            )

        # Last - if there is no 'summary' label added to image descriptor
        # we should use the value of the 'description' key and create
        # a 'summary' label with it's content. If there is even that
        # key missing - we should not add anything.
        description = self.image.label("description")

        if not self.image.label("summary") and description:
            image_labels.append(
                Label({"name": "summary", "value": description["value"]})
            )

    def _modules(self):
        """
        Returns list of modules used in all builder images as well
        as the target image.
        """

        modules = []

        for image in self.images:
            if image.modules:
                modules += [image.modules]

        return modules

    def _module_repositories(self):
        """
        Prepares list of all module repositories. This includes repositories
        defined in builder images as well as target image.
        """
        repositories = []

        for module in self._modules():
            for repo in module.repositories:
                if repo in repositories:
                    LOGGER.warning(
                        (
                            "Module repository '{0}' already added, please check your image configuration, "
                            + "skipping module repository '{0}'"
                        ).format(repo.name)
                    )
                    continue
                # If the repository already exists, skip it
                repositories.append(repo)

        return repositories

    def build_module_registry(self):
        base_dir = os.path.join(self.target, "repo")
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
        for repo in self._module_repositories():
            LOGGER.debug("Downloading module repository: '{}'".format(repo.name))
            repo.copy(base_dir)
            self.load_repository(os.path.join(base_dir, repo.target))

    def load_repository(self, repo_dir):
        for modules_dir, _, files in os.walk(repo_dir):
            if "module.yaml" in files:

                module_descriptor_path = os.path.abspath(
                    os.path.expanduser(
                        os.path.normcase(os.path.join(modules_dir, "module.yaml"))
                    )
                )

                module = Module(
                    load_descriptor(module_descriptor_path),
                    modules_dir,
                    os.path.dirname(module_descriptor_path),
                )
                LOGGER.debug(
                    "Adding module '{}', path: '{}'".format(module.name, module.path)
                )
                self._module_registry.add_module(module)

    def get_tags(self):
        return [
            "%s:%s" % (self.image["name"], self.image["version"]),
            "%s:latest" % self.image["name"],
        ]

    def copy_modules(self):
        """Prepare module to be used for Dockerfile generation.
        This means:

        1. Place module to args.target/image/modules/ directory

        """

        modules_to_install = []

        for module in self._modules():
            if module.install:
                modules_to_install += module.install

        target = os.path.join(self.target, "image", "modules")

        for module in modules_to_install:
            module = self._module_registry.get_module(
                module.name, module.version, suppress_warnings=True
            )
            LOGGER.debug(
                "Copying module '{}' required by '{}'.".format(
                    module.name, self.image.name
                )
            )

            dest = os.path.join(target, module.name)

            if not os.path.exists(dest):
                LOGGER.debug("Copying module '{}' to: '{}'".format(module.name, dest))
                shutil.copytree(module.path, dest)
            # write out the module with any overrides
            module.write(os.path.join(dest, "module.yaml"))

    def get_redhat_overrides(self):
        class RedHatOverrides(Overrides):
            def __init__(self, generator):
                super(RedHatOverrides, self).__init__({}, None)
                self._generator = generator

            @property
            def envs(self):
                return [
                    Env(
                        {
                            "name": "JBOSS_IMAGE_NAME",
                            "value": "%s" % self._generator.image["name"],
                        }
                    ),
                    Env(
                        {
                            "name": "JBOSS_IMAGE_VERSION",
                            "value": "%s" % self._generator.image["version"],
                        }
                    ),
                ]

            @property
            def labels(self):
                labels = [
                    Label(
                        {"name": "name", "value": "%s" % self._generator.image["name"]}
                    ),
                    Label(
                        {
                            "name": "version",
                            "value": "%s" % self._generator.image["version"],
                        }
                    ),
                ]
                return labels

        return RedHatOverrides(self)

    def render_dockerfile(self):
        """Renders Dockerfile to $target/image/Dockerfile"""
        LOGGER.info("Rendering Dockerfile...")

        template_file = os.path.join(
            os.path.dirname(__file__), "..", "templates", "template.jinja"
        )
        loader = FileSystemLoader(os.path.dirname(template_file))
        env = Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
        env.globals["helper"] = TemplateHelper(self._module_registry)
        env.globals["image"] = self.image
        env.globals["builders"] = self.builder_images

        template = env.get_template(os.path.basename(template_file))

        dockerfile = os.path.join(self.target, "image", "Dockerfile")
        if not os.path.exists(os.path.dirname(dockerfile)):
            os.makedirs(os.path.dirname(dockerfile))

        with open(dockerfile, "wb") as f:
            f.write(template.render(self.image).encode("utf-8"))
        LOGGER.debug("Dockerfile rendered")

    def render_help(self):
        """
        If requested, renders image help page based on the image descriptor.
        It is generated to the $target/image/help.md file.
        """

        if not self.image.help.get("add", False):
            return

        # Set default help template
        help_template_path = os.path.join(
            os.path.dirname(__file__), "..", "templates", "help.jinja"
        )

        # If custom template is requested, use it
        if self.image.get("help", {}).get("template", ""):
            help_template_path = self.image["help"]["template"]

            # If the path provided is absolute, use it
            # If it's a relative path, make it relative to the image descriptor
            if not os.path.isabs(help_template_path):
                help_template_path = os.path.join(
                    os.path.dirname(self._descriptor_path), help_template_path
                )

        LOGGER.info(
            "Rendering help.md page from template {}".format(help_template_path)
        )

        help_dirname, help_basename = os.path.split(help_template_path)
        loader = FileSystemLoader(help_dirname)
        env = Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)
        env.globals["helper"] = TemplateHelper(self._module_registry)
        env.globals["image"] = self.image
        help_template = env.get_template(help_basename)
        help_file = os.path.join(self.target, "image", "help.md")

        with open(help_file, "wb") as f:
            f.write(help_template.render(self.image).encode("utf-8"))

        LOGGER.debug("help.md rendered")

    def prepare_repositories(self):
        """Prepare repositories for build time injection."""
        if "packages" not in self.image:
            return

        if self.image.get("packages").get("content_sets"):
            LOGGER.warning(
                "The image has ContentSets repositories specified, all other repositories are removed!"
            )
            self.image["packages"]["repositories"] = []
        repos = self.image.get("packages").get("repositories", [])

        injected_repos = []

        for repo in repos:
            if self._handle_repository(repo):
                injected_repos.append(repo)

        if self.image.get("packages").get("content_sets"):
            url = self._prepare_content_sets(
                self.image.get("packages").get("content_sets")
            )
            if url:
                repo = Repository(
                    {"name": "content_sets_odcs", "url": {"repository": url}}
                )
                injected_repos.append(repo)
                self._fetch_repos = True

        if self._fetch_repos:
            for repo in injected_repos:
                repo.fetch(os.path.join(self.target, "image", "repos"))
            self.image["packages"]["repositories_injected"] = injected_repos
        else:
            self.image["packages"]["set_url"] = injected_repos

    def _prepare_content_sets(self, content_sets):
        if not content_sets:
            return False

        arch = platform.machine()

        if arch not in content_sets:
            raise CekitError(
                "There are no content_sets defined for platform '{}'!".format(arch)
            )

        repos = " ".join(content_sets[arch])

        odcs_service_type = "Fedora"
        odcs_url = "https://odcs.fedoraproject.org"

        if CONFIG.get("common", "redhat"):
            odcs_service_type = "Red Hat"
            odcs_url = "https://odcs.engineering.redhat.com"

        LOGGER.info(
            "Using {} ODCS service to create composes".format(odcs_service_type)
        )

        flags = []

        compose = (
            self.image.get("osbs", {})
            .get("configuration", {})
            .get("container", {})
            .get("compose", {})
        )

        if compose.get(Generator.ODCS_HIDDEN_REPOS_FLAG, False):
            flags.append(Generator.ODCS_HIDDEN_REPOS_FLAG)

        odcs = ODCS(odcs_url, auth_mech=AuthMech.Kerberos)

        LOGGER.debug(
            "Requesting ODCS pulp compose for '{}' repositories with '{}' flags...".format(
                repos, flags
            )
        )

        try:
            compose = odcs.new_compose(repos, "pulp", flags=flags)
        except requests.exceptions.HTTPError as ex:
            if ex.response.status_code == 401:
                LOGGER.error(
                    (
                        "You are not authorized to use {} ODCS service. "
                        "Are you sure you have a valid Kerberos session?"
                    ).format(odcs_service_type)
                )
            raise CekitError("Could not create ODCS compose", ex)

        compose_id = compose.get("id", None)

        if not compose_id:
            raise CekitError(
                "Invalid response from ODCS service: no compose id found: {}".format(
                    compose
                )
            )

        LOGGER.debug("Waiting for compose {} to finish...".format(compose_id))

        compose = odcs.wait_for_compose(compose_id, timeout=600)
        state = compose.get("state", None)

        if not state:
            raise CekitError(
                "Invalid response from ODCS service: no state found: {}".format(compose)
            )

        # State 2 is "done"
        if state != 2:
            raise CekitError("Cannot create ODCS compose: '{}'".format(compose))

        LOGGER.debug("Compose finished successfully")

        repofile = compose.get("result_repofile", None)

        if not repofile:
            raise CekitError(
                "Invalid response from ODCS service: no state_repofile key found: {}".format(
                    compose
                )
            )

        return repofile

    def _handle_repository(self, repo):
        """Process and prepares all v2 repositories.

        Args:
          repo a repository to process

        Returns True if repository file is prepared and should be injected"""

        LOGGER.debug(
            "Loading configuration for repository: '{}' from 'repositories'.".format(
                repo["name"]
            )
        )

        if "id" in repo:
            LOGGER.warning(
                "Repository '{}' is defined as plain. It must be available "
                "inside the image as Cekit will not inject it.".format(repo["name"])
            )
            return False

        if "content_sets" in repo:
            self._fetch_repos = True
            return self._prepare_content_sets(repo)

        elif "rpm" in repo:
            self._prepare_repository_rpm(repo)
            return False

        elif "url" in repo:
            return True

        return False

    def _prepare_repository_rpm(self, repo):
        pass

    def prepare_artifacts(self):
        raise NotImplementedError("Artifacts handling is not implemented")


class ModuleRegistry(object):
    def __init__(self):
        self._modules = {}
        self._defaults = {}

    def get_module(self, name, version=None, suppress_warnings=False):
        """
        Returns the module available in registry based on the name and version requested.

        If no modules are found for the requested name, an error is thrown. If version
        requirement could not be satisfied, an error is thrown too.

        If there is a version mismatch, default version is returned. See 'add_module'
        for more information how default versions are defined.

        Args:
            name (str): module name
            version (float or str): module version
            suppress_warnings: whether to suppress warnings

        Returns:
            Module object.

        Raises:
            CekitError: If a module is not found or version requirement is not satisfied
        """

        # Get all modules for specied nam
        modules = self._modules.get(name, {})

        # If there are no modules with the requested name, fail
        if not modules:
            raise CekitError(
                "There are no modules with '{}' name available".format(name)
            )

        # If there is no module version requested, get default one
        if version is None:
            default_version = self._defaults.get(name)

            if not default_version:
                raise CekitError(
                    "Internal error: default version for module '{}' could not be found, please report it".format(
                        name
                    )
                )

            default_module = self.get_module(name, default_version)

            if not suppress_warnings and len(modules) > 1:
                LOGGER.warning(
                    "Module version not specified for '{}' module, using '{}' default version".format(
                        name, default_version
                    )
                )

            return default_module

        # Finally, get the module for specified version
        module = modules.get(version)

        # If there is no such module, fail
        if not module:
            raise CekitError(
                "Module '{}' with version '{}' could not be found, available versions: {}".format(
                    name, version, ", ".join(list(modules.keys()))
                )
            )

        return module

    def add_module(self, module):
        """
        Adds provided module to registry.

        If module of the same name and version already exists in registry,
        an error is raised.

        Module registry tracks default version for a particular module name.
        For this purpose the current version is compared with what is currently defined as
        the default version. If this newer, then the default version is replaced by the
        module we currently add to registry. For version comparison the package module
        (https://packaging.pypa.io/en/latest/) is used.

        Args:
            module (Module): module object to be added to registry

        Raises:
            CekitError: when module version is not provided or when a module with the same
                name and version already exists in registry.
        """

        # If module version is not provided, fail because it is required
        if not module.version:
            raise CekitError(
                (
                    "Internal error: module '{}' does not have version specified, "
                    "we cannot add it to registry, please report it"
                ).format(module.name)
            )

        # Convert version to string, it can be float or int, or anything actually
        version = str(module.version)

        # Get all modules from registry with the name of the module we want to add
        # There can be multiple versions of the same module
        modules = self._modules.get(module.name)

        # If there are no modules for the specified name this means
        # that this is the first one, add it and set it as default
        if not modules:
            # Set it to be the default module version
            self._defaults[module.name] = version
            self._modules[module.name] = {version: module}
            return

        # If a module of specified name and version already exists in the registry - fail
        if version in modules:
            raise CekitError(
                "Module '{}' with version '{}' already exists in module registry".format(
                    module.name, version
                )
            )

        default_version = parse_version(self._defaults.get(module.name))
        current_version = parse_version(version)

        if isinstance(current_version, LegacyVersion):
            LOGGER.warning(
                (
                    "Module's '{}' version '{}' does not follow PEP 440 versioning scheme "
                    "(https://www.python.org/dev/peps/pep-0440), "
                    "we suggest follow this versioning scheme in modules"
                ).format(module.name, version)
            )

        # If current module version is never, we need to make it the new default
        if current_version > default_version:
            self._defaults[module.name] = version

        # Finally add the module to registry
        modules[version] = module



import logging
import os
import sys

# CentOS 7 does not have colorlog RPM for Python3
no_color = False
try:
    import colorlog
except ImportError:
    no_color = True


# Source: http://stackoverflow.com/questions/1383254/logging-streamhandler-and-standard-streams
# Adjusted
class SingleLevelFilter(logging.Filter):
    def __init__(self, passlevel, reject):
        self.passlevel = passlevel
        self.reject = reject

    def filter(self, record):
        if self.reject:
            return record.levelno > self.passlevel
        else:
            return record.levelno <= self.passlevel


def setup_logging(color=True):
    handler_out = logging.StreamHandler(sys.stdout)
    handler_err = logging.StreamHandler(sys.stderr)

    handler_out.addFilter(SingleLevelFilter(logging.INFO, False))
    handler_err.addFilter(SingleLevelFilter(logging.INFO, True))

    if no_color or not color or os.environ.get("NO_COLOR"):
        formatter = logging.Formatter(
            "%(asctime)s %(filename)s:%(lineno)-10s %(levelname)-5s %(message)s"
        )
    else:
        formatter = colorlog.ColoredFormatter(
            "%(log_color)s%(asctime)s %(filename)s:%(lineno)-10s %(levelname)-5s %(message)s"
        )

    handler_out.setFormatter(formatter)
    handler_err.setFormatter(formatter)

    logger = logging.getLogger("cekit")
    # Reset all handlers
    logger.handlers = []
    logger.addHandler(handler_out)
    logger.addHandler(handler_err)

    for package in ["pykwalify.rule"]:
        log = logging.getLogger(package)
        log.setLevel(logging.INFO)



import fileinput
import logging
import os
import sys
import tempfile
from collections import OrderedDict
from contextlib import closing
from typing import Dict, List
from urllib.parse import urlparse

import yaml

from cekit import crypto, version
from cekit.config import Config
from cekit.descriptor.resource import (
    Resource,
    _PlainResource,
    _PncResource,
    _UrlResource,
)
from cekit.errors import CekitError
from cekit.generator.base import Generator
from cekit.tools import copy_recursively, get_brew_url

logger = logging.getLogger("cekit")
config = Config()


class OSBSGenerator(Generator):
    def __init__(self, descriptor_path, target, overrides):
        self._wipe = True
        super(OSBSGenerator, self).__init__(descriptor_path, target, overrides)

    def init(self):
        super(OSBSGenerator, self).init()

        self._prepare_osbs_config_file(yaml.safe_dump, "container.yaml")
        self._prepare_osbs_config_file(
            lambda contents, file, **kwargs: file.write(contents), "gating.yaml"
        )

    def generate(self):
        # If extra directory exists (by default named 'osbs_extra') next to
        # the image descriptor, copy it contents to the target directory.
        #
        # https://github.com/cekit/cekit/issues/394
        copy_recursively(
            os.path.join(
                os.path.dirname(self._descriptor_path), self.image.osbs.extra_dir
            ),
            os.path.join(self.target, os.path.join("image", self.image.osbs.extra_dir)),
        )
        super(OSBSGenerator, self).generate()

    def _prepare_content_sets(self, content_sets):
        content_sets_f = os.path.join(self.target, "image", "content_sets.yml")

        if not os.path.exists(os.path.dirname(content_sets_f)):
            os.makedirs(os.path.dirname(content_sets_f))

        with open(content_sets_f, "w") as _file:
            yaml.safe_dump(content_sets, _file, default_flow_style=False)

    def _prepare_osbs_config_file(self, writer, config_file):
        config_path = os.path.join(self.target, "image", config_file)
        config_name = config_file.split(".")[0]
        all_configs = []

        if self.image.get("osbs", {}).get("configuration", {}).get(config_name):
            all_configs.append(
                self.image.get("osbs", {}).get("configuration", {}).get(config_name)
            )

        # Check all images (for multi-stage) if they contain a container definition
        for i in self.builder_images:
            if i.get("osbs", {}).get("configuration", {}).get(config_name):
                all_configs.append(
                    i.get("osbs", {}).get("configuration", {}).get(config_name, {})
                )

        if len(all_configs) > 1:
            logger.error(
                "Found multiple {} definitions ({})".format(config_name, all_configs)
            )
            raise CekitError(
                "Found multiple {} definitions ({})!".format(config_name, all_configs)
            )
        elif len(all_configs) == 0:
            return

        logger.debug(
            "Writing to {} using ident of {} with content {}".format(
                config_path, config_name, all_configs[0]
            )
        )
        if not os.path.exists(os.path.dirname(config_path)):
            os.makedirs(os.path.dirname(config_path))
        with open(config_path, "w") as _file:
            writer(all_configs[0], _file, default_flow_style=False)

    def prepare_artifacts(self):
        """Goes through artifacts section of image descriptor
        and fetches all of them
        """

        logger.info("Handling artifacts for OSBS...")
        target_dir = os.path.join(self.target, "image")
        fetch_artifacts_url: List[Dict[str, str]] = []
        fetch_artifacts_pnc = OrderedDict()
        file_comments: Dict[str, str] = {}

        fetch_domains = config.get("common", "fetch_artifact_domains")

        for image in self.images:
            for artifact in image.all_artifacts:
                logger.info(
                    "Preparing artifact '{}' (of type {})".format(
                        artifact["name"], type(artifact)
                    )
                )

                # We only want to use fetch-artifact-url if
                # 1. is type _UrlResource
                # 2. if fetch_artifact_domains configured, URL conforms to that.
                process_fetch = False
                if isinstance(artifact, _UrlResource):
                    if fetch_domains is not None:
                        fad = fetch_domains.replace(" ", "").split(",")
                        # Verify if the URL can be used in fetch-artifact-url or now
                        for d in fad:
                            u = urlparse(d)
                            logger.debug(
                                "Parsed URL '{}' and path '{}'".format(u.netloc, u.path)
                            )
                            if u.netloc + u.path in artifact["url"]:
                                process_fetch = True
                        if not process_fetch:
                            artifact["lookaside"] = True
                            logger.warning(
                                "Ignoring {} as restricted to {}".format(
                                    artifact["url"], fad
                                )
                            )
                    else:
                        # Just process all UrlResource
                        process_fetch = True

                if process_fetch:
                    intersected_hash = [
                        x for x in crypto.SUPPORTED_HASH_ALGORITHMS if x in artifact
                    ]
                    logger.debug(
                        "Found checksum markers of {}".format(intersected_hash)
                    )
                    if not intersected_hash:
                        logger.warning(
                            "No checksum supplied for {}, calculating from the remote artifact".format(
                                artifact["url"]
                            )
                        )
                        intersected_hash = ["md5"]
                        tmpfile = tempfile.NamedTemporaryFile()
                        try:
                            artifact.download_file(artifact["url"], tmpfile.name)
                            artifact["md5"] = crypto.get_sum(tmpfile.name, "md5")
                        finally:
                            tmpfile.close()

                    fetch_artifacts_url.append(
                        {
                            "url": artifact["url"],
                            "target": os.path.join(artifact["target"]),
                        }
                    )
                    for c in intersected_hash:
                        fetch_artifacts_url[len(fetch_artifacts_url) - 1].update(
                            {c: artifact[c]}
                        )
                    patch_source_url(artifact, fetch_artifacts_url)
                    if "description" in artifact:
                        file_comments[artifact["url"]] = artifact["description"]
                    logger.debug(
                        "Artifact '{}' (as URL) added to fetch-artifacts-url.yaml with contents {}".format(
                            artifact["target"],
                            fetch_artifacts_url[len(fetch_artifacts_url) - 1],
                        )
                    )
                    # OSBS by default downloads all artifacts to artifacts/<target_path>
                    artifact["target"] = os.path.join("artifacts", artifact["target"])
                elif isinstance(artifact, _PlainResource) and config.get(
                    "common", "redhat"
                ):
                    try:
                        if "md5" not in artifact:
                            logger.error(
                                "Unable to use Brew as artifact does not have md5 checksum defined"
                            )
                            raise CekitError(
                                "Unable to use Brew as artifact does not have md5 checksum defined"
                            )
                        fetch_artifacts_url.append(
                            {
                                "md5": artifact["md5"],
                                "url": get_brew_url(artifact["md5"]),
                                "target": os.path.join(artifact["target"]),
                            }
                        )
                        patch_source_url(artifact, fetch_artifacts_url)

                        logger.debug(
                            "Artifact '{}' (as plain) added to fetch-artifacts-url.yaml".format(
                                artifact["target"]
                            )
                        )
                        # OSBS by default downloads all artifacts to artifacts/<target_path>
                        artifact["target"] = os.path.join(
                            "artifacts", artifact["target"]
                        )
                    except Exception:
                        logger.warning(
                            "Plain artifact {} could not be found in Brew, trying to handle it using lookaside cache".format(
                                artifact["name"]
                            )
                        )
                        artifact.copy(target_dir)
                        # TODO: This is ugly, rewrite this!
                        artifact["lookaside"] = True

                elif isinstance(artifact, _PncResource):
                    logger.info("Handling pnc resources for {}".format(artifact))
                    build = fetch_artifacts_pnc.setdefault(artifact["pnc_build_id"], [])
                    build.append(
                        {
                            "id": artifact["pnc_artifact_id"],
                            "target": artifact["target"],
                        }
                    )
                    # OSBS by default downloads all artifacts to artifacts/<target_path>
                    artifact["target"] = os.path.join("artifacts", artifact["target"])
                    if "url" in artifact:
                        file_comments[artifact["pnc_artifact_id"]] = artifact["url"]
                else:
                    logger.debug(
                        "Copying artifact {} to {}".format(artifact, target_dir)
                    )
                    artifact.copy(target_dir)

        if fetch_artifacts_pnc:
            fetch_artifacts_file = os.path.join(
                self.target, "image", "fetch-artifacts-pnc.yaml"
            )
            pnc = {
                "metadata": {"author": "CEKit " + version.__version__},
                "builds": [
                    {"build_id": key, "artifacts": fetch_artifacts_pnc.get(key)}
                    for key in fetch_artifacts_pnc
                ],
            }
            logger.debug("Writing {} to fetch-artifacts-pnc.yaml".format(pnc))
            with open(fetch_artifacts_file, "w") as _file:
                _file.write(
                    "# Created by CEKit version {}\n".format(version.__version__)
                )
                yaml.safe_dump(pnc, _file, default_flow_style=False, sort_keys=False)
            patch_file(file_comments, fetch_artifacts_file)
        if fetch_artifacts_url:
            fetch_artifacts_file = os.path.join(
                self.target, "image", "fetch-artifacts-url.yaml"
            )
            with open(fetch_artifacts_file, "w") as _file:
                _file.write(
                    "# Created by CEKit version {}\n".format(version.__version__)
                )
                yaml.safe_dump(
                    fetch_artifacts_url,
                    _file,
                    default_flow_style=False,
                    sort_keys=False,
                )
            patch_file(file_comments, fetch_artifacts_file)
        logger.debug("Artifacts handled")


# Used to modify either the fetch-artifact or fetch-pnc files to add extra human readable information.
def patch_file(file_comments: Dict[str, str], file: str) -> None:
    # Can't use it as a context manager with plain with as that is >= 3.2
    with closing(fileinput.input(file, inplace=True)) as input_list:
        for line in input_list:
            r = [
                line.replace("\n", " # " + value + "\n")
                for (key, value) in file_comments.items()
                if key in line
            ]
            if r:
                sys.stdout.write(r[0])
            else:
                sys.stdout.write(line)


def patch_source_url(
    artifact: Resource, fetch_artifacts_url: List[Dict[str, str]]
) -> None:
    if "source-url" in artifact:
        intersected_source_hash = [
            x for x in crypto.SUPPORTED_SOURCE_HASH_ALGORITHMS if x in artifact
        ]
        if not intersected_source_hash:
            raise CekitError(
                f"Unable to add source-url for artifact {artifact} as no checksum defined"
            )
        logger.debug(
            f"Found source-url {artifact['source-url']} and checksum markers of {intersected_source_hash}"
        )
        fetch_artifacts_url[len(fetch_artifacts_url) - 1].update(
            {"source-url": artifact["source-url"]}
        )
        for c in intersected_source_hash:
            fetch_artifacts_url[len(fetch_artifacts_url) - 1].update({c: artifact[c]})



import copy

from cekit.descriptor import Image
from cekit.descriptor.image import get_image_schema

overrides_schema = get_image_schema()
overrides_schema["map"]["name"] = {"type": "str"}
overrides_schema["map"]["version"] = {"type": "text"}


class Overrides(Image):
    def __init__(self, descriptor, artifact_dir):
        self.original_descriptor = copy.deepcopy(descriptor)
        self._artifact_dir = artifact_dir
        self.path = artifact_dir
        schema = overrides_schema.copy()
        self.schema = schema
        # calling Descriptor constructor only here (we don't want Image() to mess with schema)
        super(Image, self).__init__(descriptor)

        self._prepare()



import logging
import os

import yaml

from cekit.descriptor import Descriptor
from cekit.errors import CekitError

osbs_schema = yaml.safe_load(
    """
map:
  repository: {type: any}
  configuration: {type: any}
  koji_target: {type: str}
  extra_dir: {type: str}
  extra_dir_target: {type: str}

"""
)

configuration_schema = yaml.safe_load(
    """
    map:
      container: {type: any}
      container_file: {type: str}
      gating_file: {type: str}
"""
)

repository_schema = yaml.safe_load(
    """
    map:
      name: {type: str}
      branch: {type: str}
"""
)

logger = logging.getLogger("cekit")


class Osbs(Descriptor):
    """
    Object Representing OSBS configuration

    Args:
      descriptor: dictionary object containing OSBS configuration
      descriptor_path: path to descriptor file
    """

    def __init__(self, descriptor, descriptor_path):
        self.schema = osbs_schema
        self.descriptor_path = descriptor_path
        super(Osbs, self).__init__(descriptor)

        self["configuration"] = Configuration(
            self._descriptor.get("configuration", {}), self.descriptor_path
        )

        self["repository"] = Repository(
            self._descriptor.get("repository", {}), self.descriptor_path
        )

    def merge(self, descriptor):
        if not descriptor:
            return self
        for k2, v2 in descriptor.items():
            if v2:
                self[k2] = v2
        return self

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def branch(self):
        return self.get("branch")

    @branch.setter
    def branch(self, value):
        self._descriptor["branch"] = value

    @property
    def configuration(self):
        return self.get("configuration")

    @property
    def extra_dir(self):
        return self.get("extra_dir")

    @extra_dir.setter
    def extra_dir(self, value):
        self._descriptor["extra_dir"] = value

    @property
    def extra_dir_target(self):
        return self.get("extra_dir_target")

    @extra_dir_target.setter
    def extra_dir_target(self, value):
        self._descriptor["extra_dir_target"] = value

    @property
    def koji_target(self):
        return self.get("koji_target")

    @koji_target.setter
    def koji_target(self, value):
        self._descriptor["koji_target"] = value

    @property
    def repository(self):
        return self.get("repository")


class Configuration(Descriptor):
    """Internal object representing OSBS configuration subObject

    Args:
      descriptor - yaml containing OSBS configuration"""

    def __init__(self, descriptor, descriptor_path):
        self.schema = configuration_schema
        self.descriptor_path = descriptor_path
        super(Configuration, self).__init__(descriptor)

        self._process_osbs_config_files(yaml.safe_load, "container", "container_file")
        self._process_osbs_config_files(
            lambda file: file.read(), "gating", "gating_file"
        )

        remote_source = self.get("container", {}).get("remote_source", {})
        if remote_source:
            logger.debug("Cachito definition is {}".format(remote_source))

    def _process_osbs_config_files(self, loader, text, filename):
        if text in self and filename in self:
            raise CekitError(
                "You cannot specify %s and %s together!" % (text, filename)
            )

        if filename in self:
            path = os.path.join(self.descriptor_path, self[filename])
            if not os.path.exists(path):
                raise CekitError("'%s' file not found!" % path)
            with open(path, "r") as file_:
                self[text] = loader(file_)
            del self[filename]


class Repository(Descriptor):
    def __init__(self, descriptor, descriptor_path):
        self.schema = repository_schema
        self.descriptor_path = descriptor_path
        super(Repository, self).__init__(descriptor)



import hashlib
import logging
from typing import List

logger = logging.getLogger("cekit")

SUPPORTED_HASH_ALGORITHMS: List[str] = ["sha512", "sha256", "sha1", "md5"]
SUPPORTED_SOURCE_HASH_ALGORITHMS: List[str] = [
    "source-sha1",
    "source-sha256",
    "source-md5",
]


def get_sum(target, algorithm):
    hash_function = getattr(hashlib, algorithm)()

    logger.debug("Computing {} checksum for '{}' file".format(algorithm, target))

    with open(target, "rb") as f:
        for chunk in iter(lambda: f.read(65536), b""):
            hash_function.update(chunk)
    return hash_function.hexdigest()


def check_sum(target, algorithm, expected):
    """Check that file checksum is correct
    Args:
      alg - algorithm which will be used for digest
      expected_checksum - checksum which artifact must match
    """
    logger.debug("Checking '{}' {} hash...".format(target, algorithm))

    checksum = get_sum(target, algorithm)

    if checksum.lower() != expected.lower():
        logger.error(
            "The {} computed for the '{}' file ('{}') doesn't match the '{}' value".format(
                algorithm, target, checksum, expected
            )
        )
        return False

    logger.debug("Hash is correct.")
    return True



# -*- coding: utf-8 -*-

import logging
import os
import shutil
import sys
from subprocess import CalledProcessError

import click

from cekit.config import Config, default_work_dir
from cekit.errors import CekitError
from cekit.log import setup_logging
from cekit.tools import Map
from cekit.version import __version__

LOGGER = logging.getLogger("cekit")
CONFIG = Config()


@click.group(context_settings=dict(max_content_width=100))
@click.option(
    "--descriptor",
    metavar="PATH",
    help="Path to image descriptor file.",
    default="image.yaml",
    show_default=True,
)
@click.option("-v", "--verbose", help="Enable verbose output.", is_flag=True)
@click.option("--nocolor", help="Disable color output.", is_flag=True)
@click.option(
    "--work-dir",
    metavar="PATH",
    help="Location of the working directory.",
    default=default_work_dir,
    show_default=True,
)
@click.option(
    "--config",
    metavar="PATH",
    help="Path to configuration file.",
    default=default_work_dir + "/config",
    show_default=True,
)
@click.option(
    "--redhat",
    help="Set default options for Red Hat internal infrastructure.",
    is_flag=True,
)
@click.option(
    "--target",
    metavar="PATH",
    help="Path to directory where files should be generated",
    default="target",
    show_default=True,
)
@click.version_option(message="%(version)s", version=__version__)
def cli(descriptor, verbose, nocolor, work_dir, config, redhat, target):
    """
    ABOUT

        CEKit -- Container Evolution Kit

        CEKit helps building container images from image definition files with strong focus on modularity and code reuse.

    LINKS

    \b
        Website: https://cekit.io/
        Documentation: https://docs.cekit.io/
        Issue tracker: https://github.com/cekit/cekit/issues

    EXAMPLES

        Build container image in Docker

            $ cekit build docker

        Execute tests for the example/app:1.0 image

            $ cekit test example/app:1.0
    """


@cli.group(short_help="Build container image")
@click.option(
    "--validate",
    help="Do not execute the build nor generate files, just validate image and module descriptors.",
    is_flag=True,
)
@click.option(
    "--dry-run",
    help="Do not execute the build, just generate required files.",
    is_flag=True,
)
@click.option(
    "--overrides",
    metavar="JSON",
    help="Inline overrides in JSON format.",
    multiple=True,
)
@click.option(
    "--overrides-file",
    "overrides",
    metavar="PATH",
    help="Path to overrides file in YAML format.",
    multiple=True,
)
def build(validate, dry_run, overrides):
    """
    DESCRIPTION

        Executes container image build using selected builder.

    BUILDERS

        Currently supported builders: Docker, OSBS, Podman and Buildah.

            $ cekit build BUILDER

        Run 'cekit build BUILDER --help' for more information about particular builder.

    OVERRIDES

        You can specify overrides to modify the container image build. You can read more about overrides in the documentation: https://docs.cekit.io/en/latest/overrides.html.

        Overrides can be specified inline (--overrides) or as a path to file (--overrides-file).

        Overrides can be specified multiple times.

        Please note that order matters; overrides on the right hand side take precedence! For example:

            $ cekit build --overrides '{"from": "custom/image:1.0"}' --overrides '{"from": "custom/image:2.0"}'

        Will change the 'from' key in the descriptor to 'custom/image:2.0'.
    """


@build.command(name="docker", short_help="Build using Docker engine")
@click.option("--pull", help="Always try to fetch latest base image.", is_flag=True)
@click.option(
    "--no-squash", help="Do not squash the image after build is done.", is_flag=True
)
@click.option(
    "--tag",
    "tags",
    metavar="TAG",
    help="Use specified tag to tag the image after build, can be specified multiple times.",
    multiple=True,
)
@click.option(
    "--platform",
    help="Set the ARCH of the image to the provided value instead of the architecture of the host.",
)
@click.pass_context
def build_docker(ctx, pull, no_squash, tags, platform):
    """
    DESCRIPTION

        Executes container image build locally using Docker builder.

        https://docs.docker.com/

        By default after image is built, it is squashed using the https://github.com/goldmann/docker-squash tool. You can disable it by specifying the '--no-squash' parameter.
    """
    run_build(ctx, "docker")


@build.command(name="buildah", short_help="Build using Buildah engine")
@click.option("--pull", help="Always try to fetch latest base image.", is_flag=True)
@click.option(
    "--no-squash", help="Do not squash the image after build is done.", is_flag=True
)
@click.option(
    "--tag",
    "tags",
    metavar="TAG",
    help="Use specified tag to tag the image after build, can be specified multiple times.",
    multiple=True,
)
@click.option(
    "--platform",
    help="Set the ARCH of the image to the provided value instead of the architecture of the host.",
)
@click.pass_context
def build_buildah(ctx, pull, no_squash, tags, platform):
    """
    DESCRIPTION

        Executes container image build locally using Buildah builder.

        https://buildah.io/
    """
    run_build(ctx, "buildah")


@build.command(name="podman", short_help="Build using Podman engine")
@click.option("--pull", help="Always try to fetch latest base image.", is_flag=True)
@click.option(
    "--no-squash", help="Do not squash the image after build is done.", is_flag=True
)
@click.option(
    "--tag",
    "tags",
    metavar="TAG",
    help="Use specified tag to tag the image after build, can be specified multiple times.",
    multiple=True,
)
@click.option(
    "--platform",
    help="Set the ARCH of the image to the provided value instead of the architecture of the host.",
)
@click.pass_context
def build_podman(ctx, pull, no_squash, tags, platform):
    """
    DESCRIPTION

        Executes container image build locally using Podman builder.

        https://podman.io/
    """
    run_build(ctx, "podman")


@build.command(name="osbs", short_help="Build using OSBS engine")
@click.option("--release", help="Execute a release build.", is_flag=True)
@click.option("--user", metavar="USER", help="User used to kick the build as.")
@click.option("--nowait", help="Do not wait for the task to finish.", is_flag=True)
@click.option("--stage", help="Use stage environment.", is_flag=True)
@click.option(
    "--sync-only",
    help="Generate files and sync with dist-git, but do not execute build.",
    is_flag=True,
)
@click.option(
    "--commit-message", metavar="MESSAGE", help="Custom dist-git commit message."
)
@click.option(
    "--tag",
    is_flag=False,
    flag_value="{{name}}-{{version}}",
    metavar="TAG",
    help="Use specified tag to tag the dist-git repository after build",
)
@click.option(
    "--assume-yes", "-y", help="Execute build in non-interactive mode.", is_flag=True
)
@click.pass_context
def build_osbs(
    ctx, release, user, nowait, stage, sync_only, commit_message, assume_yes, tag
):
    """
    DESCRIPTION

        Executes container image build using OSBS builder.

        https://osbs.readthedocs.io

    EXAMPLES

        Execute scratch build in OSBS

            $ cekit build osbs

        Execute regular (release) build in OSBS

            $ cekit build osbs --release

        Execute regular (release) build in OSBS with a custom commit message in dist-git

            $ cekit build osbs --release --commit-message "Release 1.0"
    """
    run_build(ctx, "osbs")


@cli.group(short_help="Execute container image tests")
@click.option("--image", help="Image to run tests against.")
@click.option(
    "--overrides",
    metavar="JSON",
    help="Inline overrides in JSON format.",
    multiple=True,
)
@click.option(
    "--overrides-file",
    "overrides",
    metavar="PATH",
    help="Path to overrides file in YAML format.",
    multiple=True,
)
def test(image, overrides):
    """
    DESCRIPTION

        Executes container image tests using selected framework.

    TEST FRAMEWORKS

        We currently support only Behave.

            $ cekit test behave

        Run 'cekit test behave --help' for more information about this particular tester.
    """


@test.command(name="behave", short_help="Run Behave tests")
@click.option(
    "--steps-url",
    help="Behave steps library.",
    default="https://github.com/cekit/behave-test-steps.git",
    show_default=True,
)
@click.option("--wip", help="Run test scenarios tagged with @wip only.", is_flag=True)
@click.option(
    "--name",
    "names",
    help="Run test scenario with the specified name, can be used specified times.",
    multiple=True,
)
@click.pass_context
def test_behave(ctx, steps_url, wip, names):
    """
    DESCRIPTION

        Execute Behave container image tests locally using Docker

        NOTE: Image to test must be available in the Docker daemon. It won't be pulled automatically!

    EXAMPLES

        Execute tests for the example/app:1.0 image

            $ cekit test --image example/app:1.0 behave

        Execute tests that are currently developed and marked with @wip

            $ cekit test --image example/app:1.0 behave --wip

        Execute specific scenario

            $ cekit test --image example/app:1.0 behave --name 'Check that product labels are correctly set'
    """

    if wip and names:
        raise click.UsageError("Parameters --name and --wip cannot be used together")

    run_test(ctx, "behave")


def prepare_params(ctx, params=None):

    if params is None:
        params = Map({})

    if ctx.parent:
        prepare_params(ctx.parent, params)

    params.update(ctx.params)

    return params


def run_command(ctx, clazz):
    params = prepare_params(ctx)
    Cekit(params).run(clazz)


def run_test(ctx, tester):
    if tester == "behave":
        from cekit.test.behave_tester import BehaveTester as tester_impl

        LOGGER.info("Using Behave tester to test the image")
    else:
        raise CekitError("Tester engine {} is not supported".format(tester))

    run_command(ctx, tester_impl)


def run_build(ctx, builder):
    if builder == "docker":
        # import is delayed until here to prevent circular import error
        from cekit.builders.docker_builder import DockerBuilder as builder_impl

        LOGGER.info("Using Docker builder to build the image")
    elif builder == "osbs":
        # import is delayed until here to prevent circular import error
        from cekit.builders.osbs import OSBSBuilder as builder_impl

        LOGGER.info("Using OSBS builder to build the image")
    elif builder == "podman":
        from cekit.builders.podman import PodmanBuilder as builder_impl

        LOGGER.info("Using Podman builder to build the image")
    elif builder == "buildah":
        from cekit.builders.buildah import BuildahBuilder as builder_impl

        LOGGER.info("Using Buildah builder to build the image")
    else:
        raise CekitError("Builder engine {} is not supported".format(builder))

    run_command(ctx, builder_impl)


class Cekit(object):
    """Main application"""

    def __init__(self, params):
        self.params = params

    def init(self):
        """Initialize logging"""
        if self.params.nocolor:
            setup_logging(False)
        else:
            setup_logging()
        if self.params.verbose:
            LOGGER.setLevel(logging.DEBUG)
        else:
            LOGGER.setLevel(logging.INFO)

        LOGGER.debug("Running version {}".format(__version__))

        if "dev" in __version__ or "rc" in __version__:
            LOGGER.warning(
                "You are running unreleased development version of CEKit, "
                "use it only at your own risk!"
            )

    def configure(self):
        """
        Prepare CEKit configuration based on config file
        and provided common parameters via CLI
        """
        LOGGER.debug("Configuring CEKit...")

        CONFIG.configure(
            self.params.config,
            {"redhat": self.params.redhat, "work_dir": self.params.work_dir},
        )

    def cleanup(self):
        """Prepares target/image directory to be regenerated."""
        directories_to_clean = [
            os.path.join(self.params.target, "image", "modules"),
            os.path.join(self.params.target, "image", "repos"),
            os.path.join(self.params.target, "repo"),
        ]

        for directory in directories_to_clean:
            if os.path.exists(directory):
                LOGGER.debug("Removing dirty directory: '{}'".format(directory))
                try:
                    shutil.rmtree(directory)
                except Exception:
                    raise CekitError("Unable to clean directory '{}'".format(directory))

    def run(self, clazz):
        """Main application entry"""

        self.init()
        self.configure()
        self.cleanup()

        command = clazz(self.params)

        try:
            command.execute()
            LOGGER.info("Finished!")
            sys.exit(0)
        except KeyboardInterrupt:
            pass
        except CalledProcessError as ex:
            LOGGER.error(ex)
            if ex.stdout:
                LOGGER.error("Command stdout is:\n{}".format(ex.stdout))
            if ex.stderr:
                LOGGER.error("Command stderr is:\n{}".format(ex.stderr))
            sys.exit(1)
        except CekitError as ex:
            if self.params.verbose:
                LOGGER.exception(ex)
            else:
                LOGGER.error(ex.message)
            sys.exit(1)


if __name__ == "__main__":
    cli()



import getpass
import logging
import os

from cekit.errors import CekitError
from cekit.tools import Chdir

try:
    from behave.__main__ import main as behave_main
except ImportError:
    pass

logger = logging.getLogger("cekit")


class BehaveTestRunner(object):
    def __init__(self, target):
        self.target = os.path.abspath(target)

    @staticmethod
    def dependencies(params=None):
        deps = {}

        deps["python-behave"] = {
            "library": "behave",
            "package": "python3-behave",
        }

        return deps

    def run(self, image, run_tags, test_names):
        """Run test suite"""
        test_path = os.path.join(self.target, "test")
        logger.debug("Running behave in '{}'.".format(test_path))
        args = [
            test_path,
            "--junit",
            "--junit-directory",
            "results",
            "--no-skipped",
            "-t",
            "~ignore",
            "-D",
            "IMAGE=%s" % image,
        ]

        if test_names:
            for name in test_names:
                args.append("--name")
                args.append("%s" % name)
        else:
            for tag in run_tags:

                # Remove anything after the colon, typically the docker tag.
                tag = tag.partition(":")[0]

                args.append("-t")
                if "/" in tag:
                    args.append("@%s,@%s" % (tag.split("/")[0], tag))
                else:
                    args.append(tag)

            # Check if we're running runtests on CI or locally
            # If we run tests locally - skip all features that
            # are marked with the @ci annotation
            if getpass.getuser() != "jenkins":
                args.append("-t")
                args.append("~ci ")

        try:
            with Chdir(os.path.join(self.target, "test")):
                logger.debug("behave args: {}".format(args))
                if behave_main(args) != 0:
                    raise CekitError(
                        "Test execution failed, please consult output above"
                    )
        except CekitError:
            raise
        except Exception:
            raise CekitError("An error occurred while executing tests")



import logging

from cekit.config import Config
from cekit.errors import CekitError
from cekit.tools import DependencyHandler

LOGGER = logging.getLogger("cekit")
CONFIG = Config()


class Command(object):
    TYPE_BUILDER = "builder"
    TYPE_TESTER = "tester"

    def __init__(self, command, command_type):
        self._command = command
        self._type = command_type

        # Initialize dependency handler
        self.dependency_handler = DependencyHandler()

        LOGGER.debug("Checking CEKit core dependencies...")
        self.dependency_handler.handle_core_dependencies()

    def execute(self):
        self.prepare()
        self.run()

    def prepare(self):
        pass

    def run(self):
        raise CekitError(
            "Command.run() method is not implemented for '{}' command and '{}' type. Please report it!".format(
                self._command, self._type
            )
        )


class Builder(Command):
    """
    Class representing generic builder - if it's instantiated it returns proper builder
    """

    def __init__(self, build_engine, params):
        self.params = params
        self.build_engine = build_engine

        self.target = self.params.target
        self.generator = None

        super(Builder, self).__init__(self.build_engine, Command.TYPE_BUILDER)

    def execute(self):
        self.prepare()
        self.before_generate()

        if self.params.validate:
            LOGGER.info(
                "The --validate parameter was specified, generation will not be performed, exiting"
            )
            return

        self.generate()

        if self.params.dry_run:
            LOGGER.info(
                "The --dry-run parameter was specified, build will not be executed, exiting"
            )
            return

        self.before_build()
        self.run()

    def prepare(self):
        if (
            self.build_engine == "docker"
            or self.build_engine == "buildah"
            or self.build_engine == "podman"
        ):
            from cekit.generator.docker import DockerGenerator as generator_impl

            LOGGER.info("Generating files for {} engine".format(self.build_engine))
        elif self.build_engine == "osbs":
            from cekit.generator.osbs import OSBSGenerator as generator_impl

            LOGGER.info("Generating files for OSBS engine")
        else:
            raise CekitError(
                "Unsupported generator type: '{}'".format(self.build_engine)
            )

        self.generator = generator_impl(
            self.params.descriptor, self.params.target, self.params.overrides
        )

        if CONFIG.get("common", "redhat"):
            # Add the redhat specific stuff after everything else
            self.generator.add_redhat_overrides()

    def before_generate(self):
        # Handle dependencies for selected generator, if any
        LOGGER.debug("Checking CEKit generate dependencies...")
        self.dependency_handler.handle(self.generator, self.params)

        self.generator.init()

    def generate(self):
        self.generator.generate()

    def before_build(self):
        LOGGER.debug("Checking CEKit build dependencies...")
        self.dependency_handler.handle(self, self.params)



import logging
import os

import yaml

from cekit.descriptor import Descriptor

logger = logging.getLogger("cekit")

volume_schema = yaml.safe_load(
    """
map:
  name: {type: str}
  path: {type: str, required: True}"""
)


class Volume(Descriptor):
    """Object representing Volume.
    If 'name' is not present its generated as basename of 'path'

    Args:
      descriptor - yaml file containing volume object
    """

    def __init__(self, descriptor):
        self.schema = volume_schema
        super(Volume, self).__init__(descriptor)
        if "name" not in self._descriptor:
            self._descriptor["name"] = os.path.basename(self._descriptor["path"])



import logging
import os
from collections.abc import MutableMapping

import yaml
from pykwalify.core import Core
from pykwalify.errors import SchemaError

from cekit.errors import CekitError

logger = logging.getLogger("cekit")


class Descriptor(MutableMapping):
    """Class serving as parent for any descriptor in cekit.

    Class implement collections.MutableMapping so it can be used as a dictionary.

    * Schema validation:
    Each class which uses this as a parent can validate its schemas defined in
    self.schemas by calling Descriptor constructor

    * Merging
    Any two Descriptor childs can be merged by invoking merge() method. Each subclass
    can define its own logic for merging by overriding this method.
    If there is any key which should not be merged, it should be appended to Descriptor.skip_merging
    list.

    args:
      descriptor - an descriptor to be represented by this class

    """

    def __init__(self, descriptor):
        self.skip_merging = []
        self._descriptor = descriptor
        self.__validate()

    def __validate(self):
        if not self.schema:
            return

        try:
            core = Core(
                source_data=self._descriptor,
                schema_data=self.schema,
                allow_assertions=True,
            )

            core.validate(raise_exception=True)
        except SchemaError as ex:
            raise CekitError(
                "Cannot validate schema: {}".format(self.__class__.__name__), ex
            )

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_data(node._descriptor)

    def write(self, path):
        directory = os.path.dirname(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(path, "w") as outfile:
            yaml.SafeDumper.add_multi_representer(Descriptor, Descriptor.to_yaml)
            yaml.dump(
                self._descriptor,
                outfile,
                default_flow_style=False,
                Dumper=yaml.SafeDumper,
            )

    def label(self, key):
        for ll in self._descriptor["labels"]:
            if ll["name"] == key:
                return ll
        return None

    def merge(self, descriptor):
        """Merges two descriptors in a way, that arrays are appended
        and duplicate values are kept
        Args:
          descriptor - a cekit descriptor
        """
        try:
            _merge_descriptors(self, descriptor)
            return self
        except KeyError as ex:
            logger.debug(ex, exc_info=True)
            raise CekitError(
                "Cannot merge descriptors, see log message for more information"
            )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self["name"] == other["name"]
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self["name"] == other["name"]
        return NotImplemented

    def __getattr__(self, name):
        """
        Make it possible to access the descriptor map as
        a property:

            o.name instead of o['name']

        If an attribute is assigned to the object and to the
        descriptor, object attribute wins and is returned.
        """

        try:
            return super(Descriptor, self).__getattr__(name)
        except AttributeError:
            pass

        if name in self._descriptor:
            return self._descriptor[name]

        return None

    def __getitem__(self, key):
        return self._descriptor[key]

    def __setitem__(self, key, item):
        self._descriptor[key] = item

    def __delitem__(self, key):
        del self._descriptor[key]

    def __iter__(self):
        return self._descriptor.__iter__()

    def __len__(self):
        return len(self._descriptor)

    def items(self):
        return self._descriptor.items()

    def __repr__(self):
        return "%s" % self._descriptor

    def get(self, k, default=None):
        return self._descriptor.get(k, default)

    def process_defaults(self):
        pass

    def remove_none_keys(self):
        if isinstance(self, Descriptor):
            _remove_none_keys(self)
        else:
            # it means it list
            for desc in self:
                _remove_none_keys(desc)


def _remove_none_keys(desc):
    for key in dict(desc.items()):
        if isinstance(desc[key], Descriptor):
            desc[key].remove_none_keys()
        if isinstance(desc[key], list):
            for d in desc[key]:
                if isinstance(d, Descriptor):
                    d.remove_none_keys()
        elif desc[key] is None:
            del desc[key]


def _merge_descriptors(desc1, desc2):
    """
    Merges two descriptors with handling embedded lists and
    descriptors.

    Args:
      desc1, desc2: descriptors to merge

    Return merged descriptor
    """
    if desc2 is None:
        return desc1
    for k2, v2 in desc2.items():
        if k2 in desc1.skip_merging:
            continue
        if k2 not in desc1:
            desc1[k2] = v2
        else:
            if isinstance(v2, Descriptor):
                desc1[k2].merge(v2)
            elif isinstance(v2, list):
                desc1[k2] = _merge_lists(desc1[k2], v2)

    return desc1


def _merge_lists(list1, list2):
    """Merges two lists handling embedded dictionaries via 'name' as a key
    In a case of simple type values are appended.

    Args:
      list1, list2 - list to merge

    Returns merged list
    """
    for v2 in reversed(list2):
        if isinstance(v2, Descriptor):
            if v2 in list1:
                v1 = list1.pop(list1.index(v2))
                list1.insert(0, v1.merge(v2))
            else:
                list1.insert(0, v2)
        elif isinstance(v2, list):
            raise CekitError("Cannot merge list of lists")
        else:
            if v2 not in list1:
                list1.insert(0, v2)

    return list1



import glob
import os
import uuid

import yaml

from cekit.config import Config
from cekit.crypto import SUPPORTED_HASH_ALGORITHMS, get_sum
from cekit.errors import CekitError

CONFIG = Config()


class ArtifactCache:
    """
    Represents Artifact cache for cekit. All cached resource are saved into cache subdirectory
    of a Cekit 'work_dir'. All files are stored by random generated uuid with an yaml file
    indexing it.
    """

    def __init__(self):
        self.cache_dir = os.path.expanduser(
            os.path.join(CONFIG.get("common", "work_dir"), "cache")
        )
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def _get_cache(self):
        cache = {}
        for index_file in glob.glob(os.path.join(self.cache_dir, "*.yaml")):
            with open(index_file, "r") as file_:
                cache[os.path.basename(index_file)] = yaml.safe_load(file_)

        return cache

    def _update_cache(self, cache_entry, artifact_id):
        index_file = os.path.join(self.cache_dir, artifact_id + ".yaml")
        tmp_cache_file = index_file + str(os.getpid())
        with open(tmp_cache_file, "w") as file_:
            yaml.safe_dump(cache_entry, file_)
            os.rename(tmp_cache_file, index_file)

    def list(self):
        return self._get_cache()

    def add(self, artifact):
        if not set(SUPPORTED_HASH_ALGORITHMS).intersection(artifact):
            raise ValueError("Cannot cache artifact without checksum")

        if self.cached(artifact):
            raise CekitError("Artifact is already cached!")

        artifact_id = str(uuid.uuid4())

        artifact_file = os.path.expanduser(os.path.join(self.cache_dir, artifact_id))
        if not os.path.exists(artifact_file):
            artifact.guarded_copy(artifact_file)

        cache_entry = {"names": [artifact["name"]], "cached_path": artifact_file}

        # We should populate the cache entry with checksums for all supported algorithms
        for alg in SUPPORTED_HASH_ALGORITHMS:
            cache_entry.update({alg: get_sum(artifact_file, alg)})

        self._update_cache(cache_entry, artifact_id)
        return artifact_id

    def delete(self, artifact_uuid):
        # check if artifact exists
        self._get_cache()
        os.remove(os.path.join(self.cache_dir, artifact_uuid))
        os.remove(os.path.join(self.cache_dir, artifact_uuid + ".yaml"))

    def get(self, artifact):
        for alg in SUPPORTED_HASH_ALGORITHMS:
            if alg in artifact:
                return self._find_artifact(alg, artifact[alg])
        raise CekitError("Artifact is not cached.")

    def _find_artifact(self, alg, chksum):
        cache = self._get_cache()
        for _, artifact in cache.items():
            if alg in artifact and artifact[alg] == chksum:
                return artifact

        raise CekitError("Artifact is not cached.")

    def cached(self, artifact):
        try:
            return self.get(artifact)
        except CekitError:
            return False



import yaml

from cekit.descriptor.base import Descriptor

port_schemas = yaml.safe_load(
    """
map:
  value: {type: int, required: True}
  protocol: {type: str}
  service: {type: str}
  expose: {type: bool}
  description: {type: str}"""
)


class Port(Descriptor):
    """Object representing Port descriptor.
    If 'name' is not present, its generated from 'value'

    args:
       descriptor - yaml object containing Port definition"""

    def __init__(self, descriptor):
        self.schema = port_schemas
        super(Port, self).__init__(descriptor)
        if "name" not in self._descriptor:
            self._descriptor["name"] = self._descriptor["value"]

    @property
    def value(self):
        return self.get("value")

    @value.setter
    def value(self, value):
        self._descriptor["value"] = value

    @property
    def protocol(self):
        return self.get("protocol")

    @protocol.setter
    def protocol(self, value):
        self._descriptor["protocol"] = value

    @property
    def service(self):
        return self.get("service")

    @service.setter
    def service(self, value):
        self._descriptor["service"] = value

    @property
    def expose(self):
        return self.get("expose")

    @expose.setter
    def expose(self, value):
        self._descriptor["expose"] = value

    @property
    def description(self):
        return self.get("description")

    @description.setter
    def description(self, value):
        self._descriptor["description"] = value



import logging
import os

import yaml

from cekit.config import Config
from cekit.descriptor import Descriptor
from cekit.descriptor.resource import create_resource
from cekit.errors import CekitError

logger = logging.getLogger("cekit")
config = Config()

packages_schema = yaml.safe_load(
    """
map:
  content_sets: {type: any}
  content_sets_file: {type: str}
  repositories:
    seq:
      - {type: any}
  install:
    seq:
      - {type: any}
  manager: {type: str, enum: ['yum', 'dnf', 'microdnf', 'apk', 'apt-get']}
  manager_flags: {type: str}"""
)


repository_schema = yaml.safe_load(
    """
map:
  name: {type: str, required: True}
  id: {type: str}
  url:
    map:
      repository: {type: str}
  rpm: {type: str}
  description: {type: str}
  odcs:
    map:
     pulp: {type: str}
  filename: {type: str}
  """
)


class Packages(Descriptor):
    """
    Object representing packages

    Args:
      descriptor - yaml containing Packages section
    """

    def __init__(self, descriptor, descriptor_path):
        self.schema = packages_schema
        self.descriptor_path = descriptor_path
        super(Packages, self).__init__(descriptor)

        # If 'content_sets' and 'content_sets_file' are defined at the same time
        if set(["content_sets", "content_sets_file"]).issubset(set(descriptor.keys())):
            raise CekitError(
                "You cannot specify 'content_sets' and 'content_sets_file' together in the packages section!"
            )

        # If the 'content_sets_file' key is set and is not None we need to read the specified
        # file and make it available in the 'content_sets' key. The 'content_sets_file' key is removed
        # afterwards.
        if descriptor.get("content_sets_file", None):
            content_sets_file = os.path.join(
                self.descriptor_path, descriptor["content_sets_file"]
            )

            if not os.path.exists(content_sets_file):
                raise CekitError("'%s' file not found!" % content_sets_file)

            with open(content_sets_file, "r") as file_:
                descriptor["content_sets"] = yaml.safe_load(file_)
            del descriptor["content_sets_file"]

        self._prepare()

    def _prepare(self):
        package_manager = self._descriptor.get("manager")
        repositories = self._descriptor.get("repositories", [])

        if (
            repositories
            and package_manager
            and package_manager not in ["yum", "dnf", "microdnf"]
        ):
            logger.warning(
                "Package manager {} does not support defining repositories, skipping all repositories".format(
                    package_manager
                )
            )
            self._descriptor["repositories"] = []
        else:
            self._descriptor["repositories"] = [Repository(x) for x in repositories]

        self._descriptor["install"] = self._descriptor.get("install", [])

    @property
    def manager(self):
        return self.get("manager")

    @property
    def manager_flags(self):
        return self.get("manager_flags")

    @property
    def repositories(self):
        return self.get("repositories")

    @property
    def install(self):
        return self.get("install")

    @property
    def content_sets(self):
        return self.get("content_sets")

    @content_sets.setter
    def content_sets(self, value):
        self._descriptor["content_sets"] = value
        self._descriptor.pop("content_sets_file", None)

    @property
    def content_sets_file(self):
        return self.get("content_sets_file")

    @content_sets_file.setter
    def content_sets_file(self, value):
        self._descriptor["content_sets_file"] = value
        self._descriptor.pop("content_sets", None)


class Repository(Descriptor):
    """Object representing package repository

    Args:
      descriptor - repository name as referenced in cekit config file
    """

    def __init__(self, descriptor):
        self.schema = repository_schema
        super(Repository, self).__init__(descriptor)

        if not (
            ("url" in descriptor)
            ^ ("odcs" in descriptor)
            ^ ("id" in descriptor)
            ^ ("rpm" in descriptor)
        ):
            raise CekitError(
                "Repository '%s' is invalid, you can use only one of "
                "['id', 'odcs', 'rpm', 'url']" % descriptor["name"]
            )

        if "filename" not in descriptor:
            descriptor["filename"] = "%s.repo" % descriptor["name"].replace(" ", "_")

        if "url" not in descriptor:
            descriptor["url"] = {}

        # we don't want to merge any of these
        self.skip_merging = ["rpm", "id", "url"]

    def fetch(self, target_dir):
        if not self._descriptor["url"]["repository"]:
            raise CekitError("Repository not defined for '{}'.".format(self.name))
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        create_resource({"url": self._descriptor["url"]["repository"]}).copy(
            os.path.join(target_dir, self._descriptor["filename"])
        )

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def description(self):
        return self.get("description")

    @description.setter
    def description(self, value):
        self._descriptor["description"] = value

    @property
    def id(self):
        return self.get("id")

    @id.setter
    def id(self, value):
        self._descriptor["id"] = value
        self._descriptor.pop("url", None)
        self._descriptor.pop("rpm", None)
        self._descriptor.pop("odcs", None)
        self._descriptor.pop("filename", None)

    @property
    def url(self):
        return self.get("url")

    @url.setter
    def url(self, value):
        self._descriptor["url"] = value
        self._descriptor.pop("id", None)
        self._descriptor.pop("rpm", None)
        self._descriptor.pop("odcs", None)
        self._descriptor.pop("filename", None)

    @property
    def rpm(self):
        return self.get("rpm")

    @rpm.setter
    def rpm(self, value):
        self._descriptor["rpm"] = value
        self._descriptor.pop("id", None)
        self._descriptor.pop("url", None)
        self._descriptor.pop("odcs", None)
        self._descriptor.pop("filename", None)

    @property
    def odcs(self):
        return self.get("odcs")

    @odcs.setter
    def odcs(self, value):
        self._descriptor["odcs"] = value
        self._descriptor.pop("id", None)
        self._descriptor.pop("url", None)
        self._descriptor.pop("rpm", None)
        self._descriptor.pop("filename", None)

    @property
    def filename(self):
        return self.get("filename")

    @filename.setter
    def filename(self, value):
        self._descriptor["filename"] = value
        self._descriptor.pop("id", None)
        self._descriptor.pop("url", None)
        self._descriptor.pop("rpm", None)
        self._descriptor.pop("odcs", None)



import yaml

from cekit.descriptor import Descriptor
from cekit.descriptor.resource import create_resource

modules_schema = yaml.safe_load(
    """
map:
  repositories:
    seq:
      - {type: any}
  install:
    seq:
      - {type: any}"""
)

install_schema = yaml.safe_load(
    """
map:
  name: {type: str, required: True}
  version: {type: str}"""
)


class Modules(Descriptor):
    def __init__(self, descriptor, path):
        self.schema = modules_schema
        super(Modules, self).__init__(descriptor)
        self._descriptor["repositories"] = [
            create_resource(r, directory=path)
            for r in self._descriptor.get("repositories", [])
        ]
        self._descriptor["install"] = [
            Install(x) for x in self._descriptor.get("install", [])
        ]

    @property
    def repositories(self):
        return self.get("repositories")

    @property
    def install(self):
        return self.get("install")


class Install(Descriptor):
    def __init__(self, descriptor):
        self.schema = install_schema
        super(Install, self).__init__(descriptor)

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def version(self):
        return self.get("version")

    @version.setter
    def version(self, value):
        self._descriptor["version"] = value



import yaml

from cekit.descriptor.base import Descriptor

label_schemas = yaml.safe_load(
    """
map:
  name: {type: str, required: True}
  value: {type: str, required: True}
  description: {type: str}
"""
)


class Label(Descriptor):
    """Object representing label descriptor.

    Args:
      descriptor - yaml object with Label
    """

    def __init__(self, descriptor):
        self.schema = label_schemas
        super(Label, self).__init__(descriptor)

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def value(self):
        return self.get("value")

    @value.setter
    def value(self, value):
        self._descriptor["value"] = value

    @property
    def description(self):
        return self.get("description")

    @description.setter
    def description(self, value):
        self._descriptor["description"] = value



import copy
import logging
from collections import OrderedDict

import yaml

import cekit
from cekit.config import Config
from cekit.descriptor import (
    Descriptor,
    Env,
    Label,
    Modules,
    Osbs,
    Packages,
    Port,
    Run,
    Volume,
)
from cekit.descriptor.resource import create_resource
from cekit.errors import CekitError
from cekit.tools import get_latest_image_version

logger = logging.getLogger("cekit")
config = Config()

_image_schema = yaml.safe_load(
    """
map:
  name: {type: str, required: True}
  version: {type: text, required: True}
  schema_version: {type: int}
  release: {type: text}
  from: {type: str}
  follow_tag: {type: str}
  description: {type: text}
  labels: {type: any}
  envs:  {type: any}
  ports: {type: any}
  run: {type: any}
  artifacts: {type: any}
  modules: {type: any}
  packages: {type: any}
  osbs: {type: any}
  volumes: {type: any}
  help:
    map:
      add: {type: bool}
      template: {type: text}"""
)


def get_image_schema():
    return copy.deepcopy(_image_schema)


class Image(Descriptor):
    def __init__(self, descriptor, artifact_dir):
        self._artifact_dir = artifact_dir
        self.path = artifact_dir
        self.schema = _image_schema.copy()
        super(Image, self).__init__(descriptor)
        self.skip_merging = ["description", "version", "name", "release"]
        self._prepare()

    def _prepare(self):
        self._descriptor["labels"] = [
            Label(x) for x in self._descriptor.get("labels", [])
        ]
        self._descriptor["envs"] = [Env(x) for x in self._descriptor.get("envs", [])]
        self._descriptor["ports"] = [Port(x) for x in self._descriptor.get("ports", [])]
        if "run" in self._descriptor:
            self._descriptor["run"] = Run(self._descriptor["run"])
        self._descriptor["artifacts"] = [
            create_resource(a, directory=self._artifact_dir)
            for a in self._descriptor.get("artifacts", [])
        ]
        self._descriptor["modules"] = Modules(
            self._descriptor.get("modules", {}), self.path
        )
        self._descriptor["packages"] = Packages(
            self._descriptor.get("packages", {}), self.path
        )
        self._descriptor["osbs"] = Osbs(self._descriptor.get("osbs", {}), self.path)
        self._descriptor["volumes"] = [
            Volume(x) for x in self._descriptor.get("volumes", [])
        ]

        # make sure image declarations override any module definitions
        self._image_overrides = {
            "artifacts": Image._to_dict(self.artifacts),
            "modules": Image._to_dict(self.modules.install),
        }
        self._all_artifacts = Image._to_dict(self.artifacts)

    def process_defaults(self):
        """Prepares default values before rendering"""
        if not self.run:
            self.run = Run({})

        # do we want to force a user?
        if "user" not in self.run:
            self.run._descriptor["user"] = cekit.DEFAULT_USER

        # Default package manager is yum
        if not self.packages.manager:
            self.packages._descriptor["manager"] = "yum"

        # Default directory for supplementary files that should be copied to dist-git directory
        if not self.osbs.extra_dir:
            self.osbs._descriptor["extra_dir"] = "osbs_extra"

        # Placing this here rather than in init as apply_image_overrides runs after that. This means
        # follow_tag is applied *after* overrides.
        if self.follow:
            if not config.get("common", "redhat"):
                raise CekitError(
                    "follow_tag annotation only supported with redhat flag"
                )
            self.base = get_latest_image_version(self.follow)

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, value):
        self._descriptor["name"] = value

    @property
    def version(self):
        return self.get("version")

    @version.setter
    def version(self, value):
        self._descriptor["version"] = value

    @property
    def release(self):
        return self.get("release")

    @release.setter
    def release(self, value):
        self._descriptor["release"] = value

    @property
    def base(self):
        return self.get("from")

    @base.setter
    def base(self, value):
        self._descriptor["from"] = value

    @property
    def follow(self):
        return self.get("follow_tag")

    @follow.setter
    def follow(self, value):
        self._descriptor["follow_tag"] = value

    @property
    def description(self):
        return self.get("description")

    @description.setter
    def description(self, value):
        self._descriptor["description"] = value

    @property
    def labels(self):
        return self.get("labels", [])

    @property
    def envs(self):
        return self.get("envs", [])

    @property
    def ports(self):
        return self.get("ports", [])

    @property
    def run(self):
        return self.get("run")

    @run.setter
    def run(self, value):
        self._descriptor["run"] = value

    @property
    def all_artifacts(self):
        return self._all_artifacts.values()

    @property
    def artifacts(self):
        return self.get("artifacts", [])

    @property
    def modules(self):
        return self.get("modules", Modules({}, self._artifact_dir))

    @property
    def packages(self):
        return self.get("packages", Packages({}, self.path))

    @property
    def osbs(self):
        return self.get("osbs")

    @osbs.setter
    def osbs(self, value):
        self._descriptor["osbs"] = value

    @property
    def volumes(self):
        return self.get("volumes", [])

    @property
    def help(self):
        return self.get("help", {})

    def apply_image_overrides(self, overrides):
        """
        Applies overrides to the image descriptor.
        """
        if not overrides:
            return
        for override in overrides:
            if override.name:
                self.name = override.name
            if override.version:
                self.version = override.version
            if override.base:
                self.base = override.base
            if override.description:
                self.description = override.description

            labels = Image._to_dict(self.labels)
            for label in override.labels:
                name = label.name
                if name in labels:
                    labels[name] = label.merge(labels[name])
                else:
                    labels[name] = label
            self._descriptor["labels"] = list(labels.values())

            envs = Image._to_dict(self.envs)
            for env in override.envs:
                name = env.name
                if name in envs:
                    envs[name] = env.merge(envs[name])
                else:
                    envs[name] = env
            self._descriptor["envs"] = list(envs.values())

            ports = Image._to_dict(self.ports)
            for port in override.ports:
                name = port.value
                if name in ports:
                    ports[name] = port.merge(ports[name])
                else:
                    ports[name] = port
            self._descriptor["ports"] = list(ports.values())

            module_repositories = Image._to_dict(self.modules.repositories)
            for repository in override.modules.repositories:
                name = repository.name
                if name in module_repositories:
                    module_repositories[name] = repository.merge(
                        module_repositories[name]
                    )
                else:
                    module_repositories[name] = repository
            self.modules._descriptor["repositories"] = list(
                module_repositories.values()
            )

            self.packages._descriptor = override.packages.merge(self.packages)

            # In case content sets are provided as null values
            # Remove the key entirely.
            # TODO: This should be handled probably at general level, for every key
            for flag in ["content_sets", "content_sets_file"]:
                if flag in override.packages and override.packages[flag] is None:
                    self.packages._descriptor.pop("content_sets", None)
                    self.packages._descriptor.pop("content_sets_file", None)

            # Merge override osbs items into self.
            self.osbs = self.osbs.merge(override.osbs)

            for package in override.packages.install:
                if package not in self.packages.install:
                    self.packages.install.append(package)

            artifact_overrides = self._image_overrides["artifacts"]
            image_artifacts = Image._to_dict(self.artifacts)
            for i, artifact in enumerate(override.artifacts):
                name = artifact.name
                # override.artifact contains override values WITH defaults.
                # override.original_descriptor contains override value WITHOUT defaults.
                # artifact_overrides contains original dictionary
                #
                # Iterating over dest / target / ...
                #   If we have _not_ supplied a target (check original_descriptor),
                #      then check artifact_overrides,
                #         otherwise use default from override.artifact
                override_without_defaults = override.original_descriptor.get(
                    "artifacts"
                )[i]
                for key in ["dest", "target", "description"]:
                    if override_without_defaults.get(key):
                        logger.debug(
                            "Key ({}) found in override as {}".format(
                                key, override_without_defaults.get(key)
                            )
                        )
                        artifact[key] = override_without_defaults.get(key)
                    elif artifact_overrides.get(name) and artifact_overrides.get(
                        name
                    ).get(key):
                        new_value = artifact_overrides.get(name).get(key)
                        logger.debug(
                            "Key ({}) found in original artifact as {}".format(
                                key, new_value
                            )
                        )
                        artifact[key] = new_value

                # collect override so we can apply it to modules
                artifact_overrides[name] = artifact
                # add it to the list of everything
                self._all_artifacts[name] = artifact
                # Apply override to image descriptor
                image_artifacts[name] = artifact
                # Sort the output as it makes it easier to view and test.
                logger.debug(
                    "Final (with override) artifact is {}".format(
                        sorted(artifact.items())
                    )
                )
            self._descriptor["artifacts"] = list(image_artifacts.values())

            module_overrides = self._image_overrides["modules"]
            image_modules = Image._to_dict(self.modules.install)
            for module in override.modules.install:
                name = module.name
                # collect override so we can apply it to modules.
                # this allows us to override module versions without affecting ordering.
                module_overrides[name] = module
                # Apply override to image descriptor
                # If the module does not exists in the original descriptor, add it there
                image_modules[name] = module
            self.modules._descriptor["install"] = list(image_modules.values())

            if override.run is not None:
                if self.run:
                    self.run = override.run.merge(self.run)
                else:
                    self.run = override.run

    def apply_module_overrides(self, module_registry):
        """
        Applies overrides to included modules.  This includes:
            Artifact definitions
            Module dependency version overrides
        Also incorporates module contributed global configuration into the image:
            Run specification
            Package repository definitions
            Required artifacts
        """
        install_list = OrderedDict()

        # index by name for easier access
        self._package_repositories = Image._to_dict(self.packages.repositories)

        # collect final 'run' value from modules
        self._module_run = Run({})

        # process the modules and integrate relevant bits into ourself
        self.process_install_list(
            self, self.modules.install, install_list, module_registry
        )

        # update ourself based on module declarations
        # final order of modules to be installed
        self.modules._descriptor["install"] = list(install_list.values())
        # all package repositories required for installing packages
        self.packages._descriptor["repositories"] = list(
            self._package_repositories.values()
        )

        # final 'run' value
        if self.run:
            self.run = self.run.merge(self._module_run)
        else:
            self.run = self._module_run

    def process_install_list(
        self, source, to_install_list, install_list, module_registry
    ) -> None:
        module_overrides = self._image_overrides["modules"]
        artifact_overrides = self._image_overrides["artifacts"]
        for to_install in to_install_list:
            logger.debug(
                "Preparing module '{}' required by '{}'.".format(
                    to_install.name, source.name
                )
            )
            override = module_overrides.get(to_install.name, None)
            if override:
                if override.version != to_install.version:
                    logger.debug(
                        "Module '{}:{}' being overridden with '{}:{}'.".format(
                            to_install.name,
                            to_install.version,
                            override.name,
                            override.version,
                        )
                    )
                # apply module override
                to_install = override

            existing = install_list.get(to_install.name, None)
            # see if we've already processed this
            if existing:
                # check for a version conflict
                if existing.version != to_install.version:
                    logger.warning(
                        "Module version inconsistency for {}: {} requested, but {} will be used.".format(
                            to_install.name, to_install.version, existing.version
                        )
                    )
                continue

            module = module_registry.get_module(to_install.name, to_install.version)
            if not module:
                raise CekitError(
                    "Could not locate module %s version %s. Please verify that it is included in one of the "
                    "specified module repositories."
                    % (to_install.name, to_install.version)
                )

            # collect artifacts and apply overrides
            module_artifacts = Image._to_dict(module.artifacts)
            for artifact in module.artifacts:
                name = artifact.name
                if name in artifact_overrides:
                    override = artifact_overrides[name]
                    self._all_artifacts[name] = override
                    module_artifacts[name] = override
                else:
                    self._all_artifacts[name] = artifact
            module._descriptor["artifacts"] = list(module_artifacts.values())

            # collect package repositories
            for repo in module.packages.repositories:
                name = repo.name
                if name not in self._package_repositories:
                    self._package_repositories[name] = repo

            # collect package manager
            if not self.packages.manager and module.packages.manager:
                logger.debug(
                    f"Applying module package manager of {module.packages.manager} to image"
                )
                self.packages._descriptor["manager"] = module.packages.manager

            # incorporate run specification contributed by module
            if module.run:
                # we're looping in order of install, so we want the current module to override whatever we have
                self._module_run = module.run.merge(self._module_run)

            # prevent circular dependencies. we'll move it to the end after processing
            install_list[to_install.name] = to_install

            # process this modules dependencies
            self.process_install_list(
                module, module.modules.install, install_list, module_registry
            )

            # move this module to the end of the list.
            install_list.pop(to_install.name)
            install_list[to_install.name] = to_install

    # helper to simplify merging lists of objects
    @classmethod
    def _to_dict(cls, named_items, key="name"):
        dictionary = OrderedDict()
        for item in named_items:
            dictionary[item[key]] = item
        return dictionary



import logging
import os

from cekit.builder import Builder
from cekit.tools import run_wrapper

LOGGER = logging.getLogger("cekit")


class BuildahBuilder(Builder):
    """This class represents buildah builder in build-using-dockerfile mode."""

    def __init__(self, params):
        super(BuildahBuilder, self).__init__("buildah", params)

    @staticmethod
    def dependencies(params=None):
        deps = {}

        deps["buildah"] = {"package": "buildah", "executable": "buildah"}

        return deps

    def run(self):
        """Build container image using buildah."""
        tags = self.params.tags
        cmd = ["/usr/bin/buildah", "build-using-dockerfile"]

        if not tags:
            tags = self.generator.get_tags()

        if not self.params.no_squash:
            cmd.append("--squash")

        if self.params.pull:
            cmd.append("--pull-always")

        if self.params.platform:
            cmd.append("--platform")
            cmd.append(self.params.platform)

        # Custom tags for the container image
        LOGGER.debug("Building image with tags: '{}'".format("', '".join(tags)))

        for tag in tags:
            cmd.extend(["-t", tag])

        LOGGER.info("Building container image...")

        cmd.append(os.path.join(self.target, "image"))

        run_wrapper(cmd, False, f"Could not run buildah {cmd}")

        LOGGER.info(
            "Image built and available under following tags: {}".format(", ".join(tags))
        )



import logging
import os
import shutil
import sys

from cekit.tools import run_wrapper

logger = logging.getLogger("cekit")


class BehaveTestCollector(object):
    def __init__(self, descriptor_dir, target_dir):
        self.collected = False
        self.descriptor_dir = os.path.abspath(descriptor_dir)
        self.target_dir = os.path.abspath(target_dir)
        self.test_dir = os.path.join(self.target_dir, "test")

        # remove old test so we can get fresh ones collected
        shutil.rmtree(self.test_dir, ignore_errors=True)
        os.makedirs(self.test_dir)

    def dependencies(self, params=None):
        deps = {}

        loader_file = os.path.join(self.test_dir, "loader.py")

        if os.path.exists(loader_file):
            try:
                sys.path.append(self.test_dir)
                from loader import StepsLoader

                sys.path.remove(self.test_dir)
                return StepsLoader.dependencies(params)
            except Exception as e:
                logger.warning(
                    "Fetching information about test dependencies failed, running tests may not be possible!"
                )
                logger.debug("Exception: {}".format(e))

        return deps

    def _fetch_steps(self, version, url):
        """Method fetches common steps"""
        logger.info("Fetching common steps from '{}'.".format(url))
        cmd = [
            "git",
            "clone",
            "--depth",
            "1",
            url,
            self.test_dir,
            "-b",
            "v%s" % version,
        ]
        run_wrapper(cmd, False, f"Could not fetch steps from {url}")

    def collect(self, version, url):
        # first clone common steps
        self._fetch_steps(version, url)
        # copy tests from repository root
        repos_dir = os.path.join(self.target_dir, "repo")
        if os.path.exists(repos_dir):
            for repo_dir in os.listdir(repos_dir):
                self._copy_tests(source=repos_dir, name=repo_dir)
        logger.debug("Collected tests from repositories roots")

        # copy tests from collected modules
        modules_dir = os.path.join(self.target_dir, "image", "modules")
        if os.path.exists(modules_dir):
            for module_dir in os.listdir(modules_dir):
                self._copy_tests(source=modules_dir, name=module_dir)
        logger.debug("Collected tests from modules")

        # copy tests from image repo
        self._copy_tests(self.descriptor_dir, "", "image")
        logger.debug("Collected tests from image")
        logger.info("Collecting finished!")

        return self.collected

    def _copy_tests(self, source, name, target_dir=""):
        for obj_name in ["steps", "features"]:
            obj_path = os.path.join(source, name, "tests", obj_name)

            if os.path.exists(obj_path):
                target = os.path.join(self.test_dir, obj_name, name, target_dir)
                logger.debug(
                    "Collecting tests from '{}' into '{}'".format(obj_path, target)
                )
                if obj_name == "features":
                    self.collected = True
                shutil.copytree(obj_path, target)



import logging
import os
import re
import traceback

from cekit.builder import Builder
from cekit.errors import CekitError

LOGGER = logging.getLogger("cekit")

# Ignore any failure on non-core modules, we will catch it later
# and suggest a solution
try:
    # Squash library
    from docker_squash.squash import Squash
except ImportError:
    pass

try:
    # Docker Python library, the new one
    import docker
    from docker.api.client import APIClient as APIClientClass
except ImportError:
    pass

try:
    # The requests library is an indirect dependency, we need to put it here
    # so that the dependency mechanism can kick in and require the docker library
    # first which will pull requests
    import requests
except ImportError:
    pass

try:
    # Docker Python library, the old one
    import docker  # noqa: F811
    from docker.client import Client as APIClientClass  # noqa: F811
except ImportError:
    pass

ANSI_ESCAPE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")


class DockerBuilder(Builder):
    """This class wraps docker build command to build and image"""

    def __init__(self, params):
        super(DockerBuilder, self).__init__("docker", params)

    @staticmethod
    def dependencies(params=None):
        deps = {}

        deps["python-docker"] = {
            "library": "docker",
            "package": "python3-docker",
            "centos7": {"package": "python36-docker"},
        }

        if params is not None and not params.no_squash:
            deps["docker-squash"] = {
                "library": "docker_squash",
                "fedora": {"package": "python3-docker-squash"},
            }

        return deps

    def _build_with_docker(self, docker_client):
        docker_args = {}
        docker_args["decode"] = True
        docker_args["path"] = os.path.join(self.target, "image")
        docker_args["pull"] = self.params.pull
        docker_args["rm"] = True
        if self.params.platform:
            docker_args["platform"] = self.params.platform

        build_log = []
        docker_layer_ids = []

        LOGGER.debug("Running Docker build: {}".format(str(docker_args)))
        try:
            stream = docker_client.build(**docker_args)

            for part in stream:
                # In case an error is returned, log the message and fail the build
                if "errorDetail" in part:
                    error_message = part.get("errorDetail", {}).get("message", "")
                    raise CekitError("Image build failed: '{}'".format(error_message))
                elif "stream" in part:
                    messages = part["stream"]
                else:
                    # We actually expect only 'stream' here.
                    # If there is something different, we ignore it.
                    # It's safe to do so because if it would be an error, we would catch it
                    # earlier. Ignored logs are related to fetching/pulling/extracting
                    # of container images.
                    continue

                # This prevents polluting CEKit log with downloading/extracting messages
                messages = ANSI_ESCAPE.sub("", messages).strip()

                for message in messages.split("\n"):
                    LOGGER.info("Docker: {}".format(message))

                build_log.append(messages)

                layer_id_match = re.search(r"^---> ([\w]{12})$", messages)

                if layer_id_match:
                    docker_layer_ids.append(layer_id_match.group(1))

        except requests.ConnectionError as ex:
            exception_chain = traceback.format_exc()
            LOGGER.debug(
                "Caught ConnectionError attempting to communicate with Docker ",
                exc_info=1,
            )

            if "PermissionError" in exception_chain:
                message = (
                    "Unable to contact docker daemon. Is it correctly setup?\n"
                    "See https://developer.fedoraproject.org/tools/docker/docker-installation.html and "
                    "http://www.projectatomic.io/blog/2015/08/why-we-dont-let-non-root-users-run-docker-in-centos-fedora-or-rhel"
                )
            elif "FileNotFoundError" in exception_chain:
                message = "Unable to contact docker daemon. Is it started?"
            else:
                message = "Unknown ConnectionError from docker ; is the daemon started and correctly setup?"

            # Work-around for python 3 code - replicate exception(...) from None
            cekit_exception = CekitError(message, ex)
            cekit_exception.__cause__ = None
            raise cekit_exception

        except Exception as ex:
            msg = "Image build failed, see logs above."
            if len(docker_layer_ids) >= 2:
                LOGGER.error(
                    "You can look inside the failed image by running "
                    "'docker run --rm -ti {} bash'".format(docker_layer_ids[-1])
                )
            if "To enable Red Hat Subscription Management repositories:" in " ".join(
                build_log
            ) and not os.path.exists(os.path.join(self.target, "image", "repos")):
                msg = (
                    "Image build failed with a yum error and you don't "
                    "have any yum repository configured, please check "
                    "your image/module descriptor for proper repository "
                    "definitions."
                )

            raise CekitError(msg, ex)

        return docker_layer_ids[-1]

    def _squash(self, docker_client, image_id):
        LOGGER.info("Squashing image {}...".format(image_id))

        squash = Squash(
            docker=docker_client,
            log=LOGGER,
            from_layer=self.generator.image["from"],
            image=image_id,
            cleanup=True,
        )
        return squash.run()

    def _tag(self, docker_client, image_id, tags):
        for tag in tags:
            if ":" in tag:
                img_repo, img_tag = tag.rsplit(":", 1)
                docker_client.tag(image_id, img_repo, tag=img_tag)
            else:
                docker_client.tag(image_id, tag)

    def _docker_client(self):
        LOGGER.debug("Preparing Docker client...")

        # Default Docker daemon connection timeout 10 minutes
        # It needs to be high enough to allow Docker daemon to export the
        # image for squashing.
        try:
            timeout = int(os.getenv("DOCKER_TIMEOUT", "600"))
        except ValueError:
            raise CekitError(
                "Provided timeout value: '{}' cannot be parsed as integer, exiting.".format(
                    os.getenv("DOCKER_TIMEOUT")
                )
            )

        if timeout <= 0:
            raise CekitError(
                "Provided timeout value needs to be greater than zero, currently: '{}', exiting.".format(
                    timeout
                )
            )

        params = {"version": "1.35"}
        params.update(docker.utils.kwargs_from_env())
        params["timeout"] = timeout

        try:
            client = APIClientClass(**params)
        except docker.errors.DockerException as e:
            LOGGER.error(
                "Could not create Docker client, please make sure that you "
                "specified valid parameters in the 'DOCKER_HOST' environment variable, "
                "examples: 'unix:///var/run/docker.sock', 'tcp://192.168.22.33:1234'"
            )
            raise CekitError("Error while creating the Docker client", e)

        if client and self._valid_docker_connection(client):
            LOGGER.debug("Docker client ready and working")
            LOGGER.debug(client.version())
            return client

        LOGGER.error(
            "Could not connect to the Docker daemon at '{}', please make sure the Docker "
            "daemon is running.".format(client.base_url)
        )

        if client.base_url.startswith("unix"):
            LOGGER.error("Please make sure the Docker socket has correct permissions.")

        if os.environ.get("DOCKER_HOST"):
            LOGGER.error(
                "If Docker daemon is running, please make sure that you specified valid "
                "parameters in the 'DOCKER_HOST' environment variable, examples: "
                "'unix:///var/run/docker.sock', 'tcp://192.168.22.33:1234'. You may "
                "also need to specify 'DOCKER_TLS_VERIFY', and 'DOCKER_CERT_PATH' "
                "environment variables."
            )

        raise CekitError("Cannot connect to Docker daemon")

    def _valid_docker_connection(self, client):
        try:
            return client.ping()
        except requests.exceptions.ConnectionError:
            pass

        return False

    def run(self):
        tags = self.params.tags

        if not tags:
            tags = self.generator.get_tags()

        LOGGER.debug("Building image with tags: '{}'".format("', '".join(tags)))
        LOGGER.info("Building container image...")

        docker_client = self._docker_client()

        # Build image
        image_id = self._build_with_docker(docker_client)

        # Squash only if --no-squash is NOT defined
        if not self.params.no_squash:
            image_id = self._squash(docker_client, image_id)

        # Tag the image
        self._tag(docker_client, image_id, tags)

        LOGGER.info(
            "Image built and available under following tags: {}".format(", ".join(tags))
        )



import configparser
import os

import yaml

default_work_dir = "~/.cekit"


class Config(object):
    """Represents Cekit configuration - behaves as Singleton"""

    cfg = {}

    @classmethod
    def configure(cls, config_path, cmdline_args):
        cls._load_cfg(config_path)
        cls._override_config(cmdline_args)

    @classmethod
    def _override_config(cls, cmdline_args):
        # Only allow command line overriding of these values if they are not the default value.
        if cmdline_args.get("redhat"):
            cls.cfg["common"]["redhat"] = cmdline_args.get("redhat")
        if (
            cmdline_args.get("work_dir")
            and cmdline_args.get("work_dir") != default_work_dir
        ):
            cls.cfg["common"]["work_dir"] = cmdline_args.get("work_dir")

    @classmethod
    def _load_cfg(cls, config_path):
        """Loads configuration from cekit config file

        params:
        config_path - path to a cekit config file (expanding user)
        """
        config_parser = configparser.ConfigParser()
        config_parser.read(os.path.expanduser(config_path))
        cls.cfg = config_parser._sections
        cls.cfg["common"] = cls.cfg.get("common", {})
        cls.cfg["common"]["work_dir"] = cls.cfg.get("common").get(
            "work_dir", default_work_dir
        )
        cls.cfg["common"]["redhat"] = yaml.safe_load(
            cls.cfg.get("common", {}).get("redhat", "False")
        )
        cls.cfg["common"]["fetch_url_domains"] = cls.cfg.get("common").get(
            "fetch_url_domains"
        )

    @classmethod
    def get(cls, *args):
        """Returns key value located by path of *args,
        None if Key doesn't exists,

        Args:
          * args - Path of key in Cekit config

        Raises:
          KeyError if section is not available."""
        value = cls.cfg
        for arg in args:
            if arg not in value and arg != args[-1]:
                raise KeyError(
                    "'%s' section doesnt exists in Cekit configuration!"
                    % "/".join(args[:-1])
                )
            value = value.get(arg)
        return value



import yaml

from cekit.descriptor import Descriptor

run_schema = yaml.safe_load(
    """
map:
  workdir: {type: str}
  user: {type: text}
  cmd:
    seq:
      - {type: str}
  entrypoint:
    seq:
      - {type: str} """
)


class Run(Descriptor):
    """Object representing Run configuration
    If 'name' is not present 'run' string is used.

    Args:
       descriptor - a yaml containing descriptor object
    """

    def __init__(self, descriptor):
        self.schema = run_schema
        super(Run, self).__init__(descriptor)
        if "name" not in self._descriptor:
            self._descriptor["name"] = "run"
        self.skip_merging = ["cmd", "entrypoint"]

    def merge(self, descriptor):
        if not descriptor:
            return self
        for k2, v2 in descriptor.items():
            if k2 not in self and v2:
                self[k2] = v2
        return self



import logging
import os

from cekit.builder import Command
from cekit.generator.behave import BehaveGenerator
from cekit.test.behave_runner import BehaveTestRunner
from cekit.test.collector import BehaveTestCollector
from cekit.version import schema_version

LOGGER = logging.getLogger("cekit")


class BehaveTester(Command):
    """
    Tester implementation for the Behave framework
    """

    def __init__(self, params):
        super(BehaveTester, self).__init__("behave", Command.TYPE_TESTER)

        self.params = params
        self.collected = False

        self.test_collector = BehaveTestCollector(
            os.path.dirname(self.params.descriptor), self.params.target
        )
        self.test_runner = BehaveTestRunner(self.params.target)

        self.generator = None

    def prepare(self):
        self.generator = BehaveGenerator(
            self.params.descriptor, self.params.target, self.params.overrides
        )

        # Handle dependencies for selected generator, if any
        LOGGER.debug("Checking CEKit generate dependencies...")
        self.dependency_handler.handle(self.generator, self.params)

        self.generator.init()

        self.generator.generate()

        # The branch to clone is determined by the schema_version in the descriptor prefixed with `v`
        # It defaults to the CEKIt current schema_version.
        self.collected = self.test_collector.collect(
            self.generator.image.get("schema_version", schema_version),
            self.params.steps_url,
        )

        if self.collected:
            # Handle test dependencies, if any
            LOGGER.debug("Checking CEKit test collector dependencies...")
            self.dependency_handler.handle(self.test_collector, self.params)
            LOGGER.debug("Checking CEKit test runner dependencies...")
            self.dependency_handler.handle(self.test_runner, self.params)

    def run(self):
        if not self.collected:
            LOGGER.warning("No test collected, test can't be run.")
            return

        test_tags = [self.generator.get_tags()[0]]

        # If wip is specified set tags to @wip
        if self.params.wip:
            test_tags = ["@wip"]

        image = self.params.image

        if not image:
            image = self.generator.get_tags()[0]

        self.test_runner.run(image, test_tags, test_names=self.params.names)



import json
import logging
import os
import shutil

from cekit.config import Config
from cekit.crypto import SUPPORTED_HASH_ALGORITHMS, check_sum
from cekit.descriptor import Descriptor
from cekit.errors import CekitError
from cekit.tools import Chdir, Map, download_file, get_brew_url, run_wrapper

logger = logging.getLogger("cekit")
config = Config()

artifact_dest = "/tmp/artifacts/"


def create_resource(descriptor, **kwargs):
    """
    Module method responsible for instantiating proper resource object
    based on the provided descriptor.

    In most cases, only descriptor is required to create the object from descriptor.
    In case additional data is required, the kwargs 'dictionary' will be checked
    if the required key exists and will be used in the object constructor.
    """

    if "image" in descriptor:
        return _ImageContentResource(descriptor)

    if "path" in descriptor:
        directory = kwargs.pop("directory")

        if not directory:
            raise CekitError(
                (
                    "Internal error: cannot instantiate PathResource: {}, directory was not provided, "
                    + "please report it: https://github.com/cekit/cekit/issues"
                ).format(descriptor)
            )

        return _PathResource(descriptor, directory)

    # PNC first so it can have an optional URL component
    if "pnc_build_id" in descriptor:
        return _PncResource(descriptor)

    if "url" in descriptor:
        return _UrlResource(descriptor)

    if "git" in descriptor:
        return _GitResource(descriptor)

    if [x for x in SUPPORTED_HASH_ALGORITHMS if x in descriptor]:
        return _PlainResource(descriptor)

    raise CekitError(
        "Unable to determine whether a URL/Git/Plain or PNC resource so '{}' is not supported".format(
            descriptor
        )
    )


class Resource(Descriptor):
    """
    Base class for handling resources.

    In most cases resources are synonym to artifacts.
    """

    CHECK_INTEGRITY = True

    def __init__(self, descriptor):
        # Schema must be provided by the implementing class
        if not self.schema:
            raise CekitError(
                "Resource '{}' has no schema defined".format(type(self).__name__)
            )

        # Includes validation
        super(Resource, self).__init__(descriptor)

        # Make sure the we have 'name' set
        self._ensure_name(descriptor)
        # Make sure the we have 'target' set
        self._ensure_target(descriptor)
        # Add a single slash at the end of the 'dest' value
        self._normalize_dest(descriptor)
        # Convert the dictionary into a Map object for easier access
        self._descriptor = self.__to_map(descriptor)

        self.skip_merging = ["md5", "sha1", "sha256", "sha512"]

        # forwarded import to prevent circular imports
        from cekit.cache.artifact import ArtifactCache

        self.cache = ArtifactCache()

    def __to_map(self, dictionary):
        """
        Convert provided dictionary, recursively, into a Map object.

        This will make it possible to access nested elements
        via properties:

                res.git.url

        instead of:

                res.git['url]
        """
        if not isinstance(dictionary, dict):
            return dictionary

        converted = Map()

        for key in dictionary:
            converted[key] = self.__to_map(dictionary[key])

        return converted

    def __eq__(self, other):
        # All subclasses of Resource are considered same object type
        if isinstance(other, Resource):
            return self["name"] == other["name"]
        return NotImplemented

    def __ne__(self, other):
        # All subclasses of Resource are considered same object type
        if isinstance(other, Resource):
            return not self["name"] == other["name"]
        return NotImplemented

    def _ensure_name(self, descriptor):
        """
        Makes sure the 'name' attribute exists.

        If it does not, a default value will be computed based on the implementation
        type of the resource class.
        """

        # If the 'name' key is present and there is a value, we have nothing to do
        if descriptor.get("name") is not None:
            return

        # Get the default value set for particular resource type
        default = self._get_default_name_value(descriptor)

        # If there is still no default, we need to fail, because 'name' is required.
        # If we ever get here, it is a bug and should be reported.
        if not default:
            raise CekitError(
                (
                    "Internal error: no value found for 'name' in '{}' artifact; unable to generate default value, "
                    + "please report it: https://github.com/cekit/cekit/issues"
                ).format(descriptor)
            )

        logger.warning(
            "No value found for 'name' in '{}' artifact; using auto-generated value of '{}'".format(
                json.dumps(descriptor, sort_keys=True), default
            )
        )

        descriptor["name"] = default

    def _ensure_target(self, descriptor):
        if descriptor.get("target") is not None:
            return

        descriptor["target"] = self._get_default_target_value(descriptor)

    def _normalize_dest(self, descriptor):
        """
        Make sure that the 'dest' value, if provided, does end with a single slash.
        """

        if descriptor.get("dest") is not None:
            descriptor["dest"] = os.path.normpath(descriptor.get("dest")) + "/"

    def _get_default_name_value(self, descriptor):
        """
        Returns default identifier value for particular class.

        This method must be overridden in classes extending Resource.
        Returned should be a string that will be be a unique identifier
        of the resource across thw whole image.
        """
        return None

    def _get_default_target_value(self, descriptor):
        return os.path.basename(descriptor.get("name"))

    def _copy_impl(self, target):
        raise NotImplementedError(
            "Implement _copy_impl() for Resource: "
            + self.__module__
            + "."
            + type(self).__name__
        )

    def copy(self, target=os.getcwd()):

        if os.path.isdir(target):
            target = os.path.join(target, self.target)

        logger.info("Copying resource '{}'...".format(self.name))

        if os.path.exists(target) and self.__verify(target):
            logger.debug("Local resource '{}' exists and is valid".format(self.name))
            return target

        cached_resource = self.cache.cached(self)

        if cached_resource:
            shutil.copy(cached_resource["cached_path"], target)
            logger.info("Using cached artifact '{}'.".format(self.name))

        else:
            try:
                self.cache.add(self)
                cached_resource = self.cache.get(self)
                shutil.copy(cached_resource["cached_path"], target)
                logger.info("Using cached artifact '{}'.".format(self.name))
            except ValueError:
                return self.guarded_copy(target)

    def guarded_copy(self, target):
        try:
            self._copy_impl(target)
        except Exception as ex:
            logger.warning(
                "Cekit is not able to fetch resource '{}' automatically. "
                "Please use cekit-cache command to add this artifact manually.".format(
                    self.name
                )
            )

            if self.description:
                logger.info(self.description)

            # exception is fatal we be logged before Cekit dies
            raise CekitError(
                "Error copying resource: '%s'. See logs for more info." % self.name, ex
            )

        if set(SUPPORTED_HASH_ALGORITHMS).intersection(self) and not self.__verify(
            target
        ):
            raise CekitError("Artifact checksum verification failed!")

        return target

    def __verify(self, target):
        """Checks all defined check_sums for an artifact"""
        if not set(SUPPORTED_HASH_ALGORITHMS).intersection(self):
            logger.debug(
                "Artifact '{}' lacks any checksum definition.".format(self.name)
            )
            return False
        if not Resource.CHECK_INTEGRITY:
            logger.info("Integrity checking disabled, skipping verification.")
            return True
        if os.path.isdir(target):
            logger.info("Target is directory, cannot verify checksum.")
            return True
        for algorithm in SUPPORTED_HASH_ALGORITHMS:
            if algorithm in self and self[algorithm]:
                if not check_sum(target, algorithm, self[algorithm]):
                    return False
        return True

    def __substitute_cache_url(self, url):
        cache = config.get("common", "cache_url")

        if not cache:
            return url

        for algorithm in SUPPORTED_HASH_ALGORITHMS:
            if algorithm in self:
                logger.debug(
                    "Using {} checksum to fetch artifacts from cacher".format(algorithm)
                )

                url = (
                    cache.replace("#filename#", self.name)
                    .replace("#algorithm#", algorithm)
                    .replace("#hash#", self[algorithm])
                )

                logger.debug("Using cache url '{}'".format(url))

        return url

    def _download_file(self, url, destination, use_cache=True):
        """Downloads a file from url and save it as destination"""
        if use_cache:
            url = self.__substitute_cache_url(url)
        if not url:
            raise CekitError(
                "Artifact %s cannot be downloaded, no URL provided" % self.name
            )
        download_file(url, destination)


class _PathResource(Resource):
    """
    Documentation: http://docs.cekit.io/en/latest/descriptor/image.html#path-artifacts
    """

    SCHEMA = {
        "map": {
            "name": {"type": "str", "desc": "Key used to identify the resource"},
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "dest": {
                "type": "str",
                "desc": "Destination directory inside of the container",
                "default": artifact_dest,
            },
            "description": {"type": "str", "desc": "Description of the resource"},
            "path": {
                "type": "str",
                "required": True,
                "desc": "Relative (suggested) or absolute path to the resource",
            },
            "md5": {"type": "str", "desc": "The md5 checksum of the resource"},
            "sha1": {"type": "str", "desc": "The sha1 checksum of the resource"},
            "sha256": {"type": "str", "desc": "The sha256 checksum of the resource"},
            "sha512": {"type": "str", "desc": "The sha512 checksum of the resource"},
        }
    }

    def __init__(self, descriptor, directory):
        self.schema = _PathResource.SCHEMA

        super(_PathResource, self).__init__(descriptor)

        path = descriptor.get("path")

        # If the path is relative it's considered relative to the directory parameter
        if not os.path.isabs(path):
            self["path"] = os.path.join(directory, path)

    def _get_default_name_value(self, descriptor):
        """
        Default identifier is the last part (most probably file name) of the URL.
        """
        return os.path.basename(descriptor.get("path"))

    def _copy_impl(self, target):
        if not os.path.exists(self.path):
            cache = config.get("common", "cache_url")

            # If cache_url is specified in Cekit configuration
            # file - try to fetch the 'path' artifact from cacher
            # even if it was not defined as 'url'.
            if cache:
                try:
                    self._download_file(self.path, target)
                    return target
                except Exception as ex:
                    logger.exception(ex)
                    raise CekitError(
                        "Could not download resource '%s' from cache" % self.name
                    )
            else:
                raise CekitError(
                    "Could not copy resource '%s', "
                    "source path does not exist. "
                    "Make sure you provided correct path" % self.name
                )

        logger.debug("Copying repository from '{}' to '{}'.".format(self.path, target))
        if os.path.isdir(self.path):
            shutil.copytree(self.path, target)
        else:
            shutil.copy2(self.path, target)
        return target


class _UrlResource(Resource):
    """
    Documentation: http://docs.cekit.io/en/latest/descriptor/image.html#url-artifacts
    """

    SCHEMA = {
        "map": {
            "name": {"type": "str", "desc": "Key used to identify the resource"},
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "dest": {
                "type": "str",
                "desc": "Destination directory inside of the container",
                "default": artifact_dest,
            },
            "description": {"type": "str", "desc": "Description of the resource"},
            "url": {
                "type": "str",
                "required": True,
                "desc": "URL where the resource can be found",
            },
            "md5": {"type": "str", "desc": "The md5 checksum of the resource"},
            "sha1": {"type": "str", "desc": "The sha1 checksum of the resource"},
            "sha256": {"type": "str", "desc": "The sha256 checksum of the resource"},
            "sha512": {"type": "str", "desc": "The sha512 checksum of the resource"},
            "source-url": {"type": "str", "desc": "The source of the resource"},
            "source-md5": {
                "type": "str",
                "desc": "The md5 checksum of the source resource",
            },
            "source-sha1": {
                "type": "str",
                "desc": "The sha1 checksum of the source resource",
            },
            "source-sha256": {
                "type": "str",
                "desc": "The sha256 checksum of the source resource",
            },
        }
    }

    def __init__(self, descriptor):
        self.schema = _UrlResource.SCHEMA

        super(_UrlResource, self).__init__(descriptor)

        # Normalize the URL
        self["url"] = descriptor.get("url").strip()

    # Avoid protected access warning
    def download_file(self, url, destination):
        return self._download_file(url, destination, False)

    def _get_default_name_value(self, descriptor):
        """
        Default identifier is the last part (most probably file name) of the URL.
        """
        return os.path.basename(descriptor.get("url"))

    def _copy_impl(self, target):
        try:
            self._download_file(self.url, target)
        except Exception:
            logger.debug(
                "Cannot hit artifact: '{}' via cache, trying directly.".format(
                    self.name
                )
            )
            self._download_file(self.url, target, use_cache=False)
        return target


class _GitResource(Resource):

    SCHEMA = {
        "map": {
            "name": {"type": "str", "desc": "Key used to identify the resource"},
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "description": {"type": "str", "desc": "Description of the resource"},
            "git": {
                "required": True,
                "map": {
                    "url": {
                        "type": "str",
                        "required": True,
                        "desc": "URL of the repository",
                    },
                    "ref": {
                        "type": "str",
                        "required": True,
                        "desc": "Reference to check out; could be branch, tag, etc",
                    },
                },
            },
        }
    }

    def __init__(self, descriptor):
        self.schema = _GitResource.SCHEMA

        super(_GitResource, self).__init__(descriptor)

    def _get_default_name_value(self, descriptor):
        return os.path.basename(descriptor.get("git", {}).get("url")).split(".", 1)[0]

    def _copy_impl(self, target):
        cmd = ["git", "clone", self.git.url, target]
        run_wrapper(cmd, False, f"Could not clone from {self.git.url}")

        with Chdir(target):
            cmd = ["git", "checkout", self.git.ref]
            run_wrapper(cmd, False, f"Could not checkout from {self.git.ref}")

        return target


class _PlainResource(Resource):
    """
    Documentation: http://docs.cekit.io/en/latest/descriptor/image.html#plain-artifacts
    """

    SCHEMA = {
        "map": {
            "name": {
                "type": "str",
                "required": True,
                "desc": "Key used to identify the resource",
            },
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "dest": {
                "type": "str",
                "desc": "Destination directory inside of the container",
                "default": artifact_dest,
            },
            "description": {"type": "str", "desc": "Description of the resource"},
            "md5": {"type": "str", "desc": "The md5 checksum of the resource"},
            "sha1": {"type": "str", "desc": "The sha1 checksum of the resource"},
            "sha256": {"type": "str", "desc": "The sha256 checksum of the resource"},
            "sha512": {"type": "str", "desc": "The sha512 checksum of the resource"},
            "source-url": {"type": "str", "desc": "The source of the resource"},
            "source-md5": {
                "type": "str",
                "desc": "The md5 checksum of the source resource",
            },
            "source-sha1": {
                "type": "str",
                "desc": "The sha1 checksum of the source resource",
            },
            "source-sha256": {
                "type": "str",
                "desc": "The sha256 checksum of the source resource",
            },
        }
    }

    def __init__(self, descriptor):
        self.schema = _PlainResource.SCHEMA

        super(_PlainResource, self).__init__(descriptor)

    def _copy_impl(self, target):
        # First of all try to download the file using cacher if specified
        if config.get("common", "cache_url"):
            try:
                self._download_file(None, target)
                return target
            except Exception as e:
                logger.debug(str(e))
                logger.warning(
                    "Could not download '{}' artifact using cacher".format(self.name)
                )

        # Next option is to download it from Brew directly but only if the md5 checksum
        # is provided and we are running with the --redhat switch
        if self.md5 and config.get("common", "redhat"):
            logger.debug(
                "Trying to download artifact '{}' from Brew directly".format(self.name)
            )

            try:
                # Generate the URL
                url = get_brew_url(self.md5)
                # Use the URL to download the file
                self._download_file(url, target, use_cache=False)
                return target
            except Exception as e:
                logger.debug(str(e))
                logger.warning(
                    "Could not download artifact '{}' from Brew".format(self.name)
                )

        raise CekitError("Artifact {} could not be found".format(self.name))


class _ImageContentResource(Resource):
    """
    Class to cover artifacts that should be fetched from images.

    Main purpose of this type of resources are artifacts built as
    part of multi-stage builds. Other use case is where the artifact
    should be fetched from an already built image.

    Such a resource is represented in Dockerfile as:

        COPY --from=[IMAGE] [PATH] /tmp/artifacts/[NAME]

    Due to the nature of such resources, we're not able to validate
    checksums of such resources.
    """

    SCHEMA = {
        "map": {
            "name": {"type": "str", "desc": "Key used to identify the resource"},
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "dest": {
                "type": "str",
                "desc": "Destination directory inside of the container",
                "default": "/tmp/artifacts/",
            },
            "description": {"type": "str", "desc": "Description of the resource"},
            "image": {
                "type": "str",
                "required": True,
                "desc": "Name of the image which holds the resource",
            },
            "path": {
                "type": "str",
                "required": True,
                "desc": "Path in the image under which the resource can be found",
            },
        }
    }

    def __init__(self, descriptor):
        self.schema = _ImageContentResource.SCHEMA

        super(_ImageContentResource, self).__init__(descriptor)

    def _get_default_name_value(self, descriptor):
        """
        Default identifier is the file name of the resource inside of the image.
        """
        return os.path.basename(descriptor.get("path"))

    def _copy_impl(self, target):
        """
        For stage artifacts, there is nothing to copy, because the artifact is located
        in an image that should be built in earlier stage of the image build process.
        """
        return target


class _PncResource(Resource):
    """
    Documentation: http://docs.cekit.io/en/latest/descriptor/image.html#pnc-artifacts
    """

    SCHEMA = {
        "map": {
            "name": {"type": "str", "desc": "Key used to identify the resource"},
            "target": {"type": "str", "desc": "Target file name for the resource"},
            "dest": {
                "type": "str",
                "desc": "Destination directory inside of the container",
                "default": artifact_dest,
            },
            "description": {"type": "str", "desc": "Description of the resource"},
            "pnc_artifact_id": {
                "type": "str",
                "required": True,
                "desc": "The ID of the artifact",
            },
            "pnc_build_id": {
                "type": "str",
                "required": True,
                "desc": "The ID of the build",
            },
            "url": {
                "type": "str",
                "desc": "Optional URL where the resource can be found",
            },
        }
    }

    def __init__(self, descriptor):
        self.schema = _PncResource.SCHEMA

        super(_PncResource, self).__init__(descriptor)

    def _get_default_name_value(self, descriptor):
        """
        Default identifier is the target file name.
        """
        return descriptor.get("target")

    def _copy_impl(self, target):
        """
        For PNC artifacts there is nothing to copy as the artifact is held remotely on PNC.
        """
        return target



import os


class TemplateHelper(object):

    SUPPORTED_PACKAGE_MANAGERS = ["yum", "dnf", "microdnf", "apk", "apt-get"]

    def __init__(self, module_registry):
        self._module_registry = module_registry

    def module(self, to_install):
        return self._module_registry.get_module(
            to_install.name, to_install.version, suppress_warnings=True
        )

    def packages_to_install(self, image):
        """
        Method that returns list of packages to be installed by any of
        modules or directly in the image
        """
        all_modules = self.modules(image)

        packages = []

        for module in all_modules:
            if "packages" in module and "install" in module.packages:
                packages += module.packages.install

        return packages

    def modules(self, image):
        all_modules = []

        if "modules" in image and "install" in image.modules:
            all_modules += [self.module(m) for m in image.modules.install]

        all_modules.append(image)

        return all_modules

    def filename(self, source):
        """Simple helper to return the file specified name"""

        target = source.get("target")

        if target:
            return target

        return os.path.basename(source["artifact"])

    def cmd(self, arr):
        """Generates array of commands that could be used like this:
        CMD {{ helper.cmd(cmd) }}
        """

        ret = []
        for cmd in arr:
            ret.append('"%s"' % cmd)
        return "[%s]" % ", ".join(ret)

    def all_envs(self, image):
        envs = []
        for module in self.modules(image):
            envs += module.envs

        return envs

    def all_labels(self, image):
        labels = []
        for module in self.modules(image):
            labels += module.labels

        return labels

    def ports(self, available_ports):
        """
        Combines all ports that should be added to the
        Dockerfile into one array
        """

        port_list = []

        for p in available_ports:
            if p.get("expose", True):
                port_list.append(p.get("value"))

        return port_list

    def cachito(self, image):
        if (
            image.get("osbs", {})
            .configuration.get("container", {})
            .get("remote_source")
        ):
            return True

    def extra_dir(self, image):
        return image.get("osbs", {}).extra_dir

    def extra_dir_target(self, image):
        return image.get("osbs", {}).extra_dir_target

    def package_manager_flags(self, pkg_mgr, pkg_mgr_flags):
        if pkg_mgr_flags is not None:
            # Using None check to allow a definition of "manager_flags: ''" to override the default values.
            return pkg_mgr_flags
        elif "apk" in pkg_mgr:
            return ""
        elif "apt-get" in pkg_mgr:
            #
            # This is a HACK...
            #
            # Debian based apt-get needs an *update* step
            # *before* its "install" step...
            #
            # We really *should* add an additional step to the
            # main template repo_install and pkg_install macros
            #
            # However this works at the moment...
            #
            return "update && apt-get --no-install-recommends"

        default = "--setopt=tsflags=nodocs"
        if "microdnf" in pkg_mgr:
            return "--setopt=install_weak_deps=0 " + default
        else:
            return default

    def package_manager_install(self, pkg_mgr):
        if "apk" in pkg_mgr:
            return "add"
        else:
            return "install -y"

    def package_manager_query(self, pkg_mgr):
        if "apk" in pkg_mgr:
            return "apk info -e"
        elif "apt-get" in pkg_mgr:
            return "dpkg-query --list"
        else:
            return "rpm -q"



from cekit.descriptor import Execute, Image
from cekit.descriptor.image import get_image_schema

overrides_schema = get_image_schema()
overrides_schema["map"]["execute"] = {"type": "any"}


class Module(Image):
    """Represents a module.

    Constructor arguments:
    descriptor_path: A path to module descriptor file.
    """

    def __init__(self, descriptor, path, artifact_dir):
        self._artifact_dir = artifact_dir
        self.path = path
        self.schema = overrides_schema.copy()
        # calling Descriptor constructor only here (we don't want Image() to mess with schema)
        super(Image, self).__init__(descriptor)
        self.skip_merging = ["description", "version", "name", "release", "help"]

        self._prepare()
        self.name = self._descriptor["name"]
        self._descriptor["execute"] = [
            Execute(x, self.name) for x in self._descriptor.get("execute", [])
        ]

    @property
    def execute(self):
        return self.get("execute")



#!/usr/bin/python

import codecs

from setuptools import find_packages, setup

from cekit.version import __version__

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="cekit",
    version=__version__,
    packages=find_packages(exclude=["tests"]),
    package_data={
        "cekit.templates": ["*.jinja"],
        "cekit.schema": ["*.yaml"],
    },
    url="https://github.com/cekit/cekit",
    download_url="https://github.com/cekit/cekit/archive/%s.tar.gz" % __version__,
    author="CEKit team",
    author_email="cekit@cekit.io",
    description="Container image creation tool",
    long_description=codecs.open("README.rst", encoding="utf-8").read(),
    license="MIT",
    entry_points={
        "console_scripts": ["cekit=cekit.cli:cli", "cekit-cache=cekit.cache.cli:cli"],
    },
    tests_require=["mock"],
    install_requires=requirements,
)



import logging
import shutil
import sys

import click

from cekit.cache.artifact import ArtifactCache
from cekit.config import Config
from cekit.crypto import SUPPORTED_HASH_ALGORITHMS
from cekit.descriptor.resource import create_resource
from cekit.log import setup_logging
from cekit.tools import Map
from cekit.version import __version__

setup_logging()
LOGGER = logging.getLogger("cekit")
CONFIG = Config()


@click.group(context_settings=dict(max_content_width=100))
@click.option("-v", "--verbose", help="Enable verbose output.", is_flag=True)
@click.option(
    "--config",
    metavar="PATH",
    help="Path to configuration file.",
    default="~/.cekit/config",
    show_default=True,
)
@click.option(
    "--work-dir",
    metavar="PATH",
    help="Location of the working directory.",
    default="~/.cekit",
    show_default=True,
)
@click.version_option(message="%(version)s", version=__version__)
def cli(config, verbose, work_dir):
    pass


@cli.command(name="ls", short_help="List cached artifacts")
def ls():
    CacheCli.prepare().ls()


@cli.command(name="add", short_help="Add artifact to cache")
@click.argument("location", metavar="LOCATION")
@click.option("--md5", metavar="CHECKSUM", help="The md5 checksum of the artifact.")
@click.option("--sha1", metavar="CHECKSUM", help="The sha1 checksum of the artifact.")
@click.option(
    "--sha256", metavar="CHECKSUM", help="The sha256 checksum of the artifact."
)
@click.option(
    "--sha512", metavar="CHECKSUM", help="The sha512 checksum of the artifact."
)
def add(location, md5, sha1, sha256, sha512):
    if not (md5 or sha1 or sha256 or sha512):
        raise click.UsageError("At least one checksum must be provided")

    CacheCli.prepare().add(location, md5, sha1, sha256, sha512)


@cli.command(name="rm", short_help="Remove artifact from cache")
@click.argument("uuid", metavar="UUID")
def rm(uuid):
    CacheCli.prepare().rm(uuid)


@cli.command(name="clear", short_help="Remove all artifacts from the cache")
def clear():
    CacheCli.prepare().clear()


class CacheCli:
    @staticmethod
    def prepare():
        """Returns an initialized object of CacheCli"""
        return CacheCli(Map(click.get_current_context().parent.params))

    def __init__(self, args):

        # TODO: logging is used only when adding the artifact, we need to find out if it would be possible to do it better
        if args.verbose:
            LOGGER.setLevel(logging.DEBUG)
        else:
            LOGGER.setLevel(logging.INFO)

        CONFIG.configure(args.config, {"work_dir": args.work_dir})

    def add(self, location, md5, sha1, sha256, sha512):
        artifact_cache = ArtifactCache()

        resource = {}
        resource["url"] = location

        if md5:
            resource["md5"] = md5

        if sha1:
            resource["sha1"] = sha1

        if sha256:
            resource["sha256"] = sha256

        if sha512:
            resource["sha512"] = sha512

        artifact = create_resource(resource)

        cached = artifact_cache.cached(artifact)

        if cached:
            click.echo("Artifact {} is already cached!".format(location))
            sys.exit(0)

        try:
            artifact_id = artifact_cache.add(artifact)
            click.echo(
                "Artifact {} cached with UUID '{}'".format(location, artifact_id)
            )
        except Exception as ex:
            click.secho(
                "Cannot cache artifact {}: {}".format(location, str(ex)), fg="red"
            )
            sys.exit(1)

    def ls(self):
        artifact_cache = ArtifactCache()
        artifacts = artifact_cache.list()
        if artifacts:
            for artifact_filename, artifact in artifacts.items():
                click.echo(
                    "\n{}:".format(
                        click.style(
                            artifact_filename.split(".")[0], fg="green", bold=True
                        )
                    )
                )
                for alg in SUPPORTED_HASH_ALGORITHMS:
                    if alg in artifact and artifact[alg]:
                        click.echo(
                            "  {}: {}".format(
                                click.style(alg, bold=True), artifact[alg]
                            )
                        )

                if artifact["names"]:
                    click.echo("  {}:".format(click.style("names", bold=True)))
                    for name in artifact["names"]:
                        click.echo("    - %s" % name)
        else:
            click.echo("No artifacts cached!")

    def rm(self, uuid):
        artifact_cache = ArtifactCache()

        try:
            artifact_cache.delete(uuid)
            click.echo("Artifact with UUID '{}' removed".format(uuid))
        except Exception:
            click.secho(
                "Artifact with UUID '{}' doesn't exists in the cache".format(uuid),
                fg="yellow",
            )
            sys.exit(1)

    def clear(self):
        """
        Removes the artifact cache directory with all artifacts.

        Use with caution!
        """
        artifact_cache = ArtifactCache()

        if not click.confirm(
            "Are you sure to remove all artifacts from cache?", show_default=True
        ):
            return

        try:
            shutil.rmtree(artifact_cache.cache_dir)
            click.echo("Artifact cache cleared!")
        except Exception:
            click.secho(
                "An error occurred while removing the artifact cache directory '{}'".format(
                    artifact_cache.cache_dir
                ),
                fg="red",
            )
            sys.exit(1)


if __name__ == "__main__":
    cli()
